import inspect
from typing import Annotated, Iterable, override

import atlassian
import fastmcp.tools
import httpx
import pydantic

from bir_mcp.atlassian.utils import (
    generate_directed_linked_issues,
    infer_jira_field_representation,
    normalize_line_breaks,
)
from bir_mcp.core import AhttpxMixin, BaseMcp, build_readonly_tools, build_write_tools
from bir_mcp.utils import (
    araise_for_status,
    build_url,
    filter_dict_by_keys,
    llm_friendly_http_request,
    prepend_url_path_prefix_if_not_present,
    to_maybe_ssl_context,
)

IssueKeyType = Annotated[
    str,
    pydantic.Field(
        description="The Jira issue key in the format {jira_project_key}-{issue_sequntial_number}, for example ABC-123"
    ),
]


class Atlassian(BaseMcp, AhttpxMixin):
    def __init__(
        self,
        token: str,
        url: str = "https://jira-support.kapitalbank.az",
        api_version: int = 2,
        http_timeout_seconds: int = 10,
        backoff_and_retry: bool = True,
        retry_status_codes: Iterable[int] = (413, 429, 503, 504),
        max_backoff_seconds: int = 10,
        max_backoff_retries: int = 3,
        timezone: str = "UTC",
        ssl_verify: bool | str = True,
    ):
        super().__init__(timezone=timezone)
        self.jira = atlassian.Jira(
            url=url,
            token=token,
            api_version=api_version,
            verify_ssl=bool(ssl_verify),
            timeout=http_timeout_seconds,
            backoff_and_retry=backoff_and_retry,
            retry_status_codes=retry_status_codes,
            max_backoff_retries=max_backoff_retries,
            max_backoff_seconds=max_backoff_seconds,
        )
        headers = {
            "Authorization": f"Bearer {token}",
            "Accept": "application/json",
            "Content-Type": "application/json",
        }
        self.ahttpx = httpx.AsyncClient(
            base_url=url,
            headers=headers,
            event_hooks={"response": [araise_for_status]},
            timeout=http_timeout_seconds,
            verify=to_maybe_ssl_context(ssl_verify),
        )
        self.api_version = api_version
        # anon_codec: AnonCodec
        # anon_codec.anonymize

    @override
    def get_tag(self):
        return "atlassian"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="Bir Atlassian MCP server",
            instructions=inspect.cleandoc("""
                Atlassian related tools, such as Jira, Jira Service Desk, Confluence, etc.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        read_tools = [
            self.get_raw_jira_issue_fields,  # TODO: need to use anon_codec.anonymize on its output.
            self.get_allowed_issue_types_for_project,
            self.search_users,
            self.get_jira_issue_overview,
        ]
        write_tools = [
            self.create_jira_issue,
            self.jira_http_request,
        ]
        tools = build_readonly_tools(read_tools, max_output_length=max_output_length, tags=tags)
        tools.extend(build_write_tools(write_tools, max_output_length=max_output_length, tags=tags))
        return tools

    def build_issue_url(self, issue_key: str) -> str:
        return f"{self.jira.url}/browse/{issue_key}"

    def prepend_api(self, path: str) -> str:
        return prepend_url_path_prefix_if_not_present(f"rest/api/{self.api_version}", path)

    async def get_allowed_issue_types_for_project(self, project_key: str) -> dict:
        """
        Fetch all Jira issue types allowed in the project, grouped into tasks and subtasks,
        with optional descriptions.
        """
        project = await self.ahttpx_get(f"project/{project_key}")
        project = project.json()
        issue_types = project["issueTypes"]
        tasks = {}
        subtasks = {}
        for issue_type in issue_types:
            filtered_issue_type = {issue_type["name"]: issue_type["description"]}
            if issue_type["subtask"]:
                subtasks |= filtered_issue_type
            else:
                tasks |= filtered_issue_type

        return dict(tasks=tasks, subtasks=subtasks)

    async def jira_http_request(
        self,
        path: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The path to the Jira API endpoint, relative to the base URL.
                    Should not include the /rest/api/version prefix.
                """)
            ),
        ],
        http_method: str = "GET",
        json_query_params: Annotated[
            str | None,
            pydantic.Field(description="The url query parameters dict, serialized to JSON."),
        ] = None,
        content: Annotated[
            str | dict | None,
            pydantic.Field(description="The request body."),
        ] = None,
    ) -> dict:
        """
        Performs a HTTP request to the Jira REST API. Can be used when no specialized tool is available for the task.
        The arguments are passed to a httpx.Client.request method, with base_url set to local
        Jira instance and API key set in headers, so the result looks like this:
        return client.request(method=http_method, url=path, params=json.loads(json_query_params), content=content).json()
        """
        data = await llm_friendly_http_request(
            client=self.ahttpx,
            url=self.prepend_api(path),
            http_method=http_method,
            json_query_params=json_query_params,
            content=content,
            ensure_dict_output=True,
        )
        return data

    async def get_raw_jira_issue_fields(
        self,
        issue_key: IssueKeyType,
        issue_fields: Annotated[
            Iterable[str],
            pydantic.Field(description="The fields to fetch from the Jira issue"),
        ] = ("summary", "description"),
    ) -> dict:
        """Fetch raw issue fields for given Jira issue."""
        response = await self.ahttpx_get(f"issue/{issue_key}", fields=",".join(issue_fields))
        response = response.json()
        fields = response["fields"]
        return fields

    async def get_jira_issue_overview(self, issue_key: IssueKeyType) -> dict:
        """
        Fetch a preformatted set of general Jira issue fields and attributes,
        such as summary, description, comments, issue type, parent, subtasks,
        issue links, priority.
        """
        fields = [
            "summary",
            "description",
            "comment",
            "issuetype",
            "parent",
            "subtasks",
            "issuelinks",
            "priority",
        ]
        response = await self.ahttpx_get(f"issue/{issue_key}", fields=",".join(fields))
        fields = response.json()["fields"]
        fields = {k: infer_jira_field_representation(v) for k, v in fields.items()}

        comments = fields.pop("comment")["comments"]
        comments = sorted(comments, key=lambda c: c["created"])
        comments = [f"{c['author']['displayName']}: {c['body']}" for c in comments]
        fields["comments"] = comments

        fields["issuelinks"] = [
            f"{directed_link}: {linked_issue['key']}"
            for directed_link, linked_issue in generate_directed_linked_issues(fields)
        ]

        fields = {
            key: normalize_line_breaks(value) if isinstance(value, str) else value
            for key, value in fields.items()
        }
        fields["jira_issue_url"] = build_url(self.jira.url, "browse", issue_key)
        return fields

    async def create_jira_issue(
        self,
        project_key: Annotated[
            str, pydantic.Field(description="The key of a Jira project to create an issue in.")
        ],
        description: str,
        summary: str,
        issue_type: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc(f"""
                    The name of a Jira issue type to create, for example Development Task, Bug, Kanban Feature.
                    Note that each project has its own set of allowed issue types, for details refer to the 
                    "{get_allowed_issue_types_for_project.__name__}" tool.
                """),
            ),
        ],
        assignee: Annotated[
            str | None,
            pydantic.Field(description="The name of a Jira user to assign the issue to."),
        ] = None,
        additional_fields: Annotated[
            dict | None, pydantic.Field(description="Additional fields to set on the Jira issue.")
        ] = None,
    ) -> dict:
        """Create a new Jira issue and get its key."""
        fields = additional_fields or {}
        fields |= {
            "summary": summary,
            "description": description,
            "project": {"key": project_key},
            "issuetype": {"name": issue_type},
        }
        if assignee:
            fields["assignee"] = {"name": assignee}

        response = await self.ahttpx_post("issue", json={"fields": fields})
        issue = response.json()
        issue_key = issue["key"]
        created_issue = {
            "created_issue_key": issue_key,
            "url": self.build_issue_url(issue_key),
        }
        return created_issue

    async def search_users(
        self,
        query: Annotated[
            str,
            pydantic.Field(
                description="The query to search users by a non-case-sensitive match of username, name or email."
            ),
        ],
    ) -> dict:
        """Search for Jira users, primarily usefull to find an exact user name."""
        # Use async HTTP GET to search users
        response = await self.ahttpx_get("user/search", username=query)
        users = response.json()
        users = [
            filter_dict_by_keys(user, ["name", "displayName", "emailAddress"]) for user in users
        ]
        users = dict(users=users)
        return users



import inspect

from bir_mcp.atlassian import Atlassian
from bir_mcp.git_lab.sonarqube import SonarQube
from bir_mcp.git_lab.tools import GitLabTools
from bir_mcp.utils import wrap_with_xml_tag


def get_prompts() -> list[callable]:
    prompts = [
        get_merge_request_review_prompt,
        get_feature_development_prompt,
    ]
    return prompts


def get_merge_request_review_prompt(merge_request_url: str | None = None) -> str:
    """A prompt for reviewing a GitLab merge request."""
    merge_request = f'The merge request url is "{merge_request_url}".' if merge_request_url else ""
    prompt = inspect.cleandoc(f"""
        Your job is to conduct a thorough review of a GitLab merge request.

        Follow these steps:
        - Fetch the merge request overview using the {GitLabTools.get_merge_request_overview_from_url.__name__} tool.
        - Review the pipeline jobs, their statuses and traces, explain the meaning and value of the jobs, consider whether 
        they sufficiently verified the code quality and security.
        - Inspect the SonarQube data for the GitLab project's source and target branches using the 
        {SonarQube.get_project_quality_gates_status.__name__} and {SonarQube.get_main_project_metrics.__name__} tools.
        Pay attention to the analysis age, if it is older than 1 month, consider it stale. 
        - Inspect the first lines of the diffs in the merge request overview. Explain the purpose of the changes. 
        Evaluate the quality of the code, check for any potential bugs and security issues. If the first lines are
        not enough, use the {GitLabTools.get_merge_request_file_diffs.__name__} tool for full diffs and the 
        {GitLabTools.get_file_content.__name__} tool for the full file content in the source and target branches.
        Do not call the {GitLabTools.get_merge_request_file_diffs.__name__} tool with individual files, but rather
        collect all file paths that you need in a single call.
        - If the merge request overview returned any related Jira issues, use the {Atlassian.get_jira_issue_overview.__name__}
        tool to fetch the issues. Ensure that the merge changes are related to the issues, that the issues have
        appropriate status and priority, that the issues include mentions of testing and have the necessary approvals.
        - Prepare the review summary, with purpose and explanation of the changes, findings from previous steps, 
        and the verdict, whether the changes are ready to be merged, if not, provide reasons and suggest next actions.

        {merge_request}
    """)
    # - If available, use the build HTML page tool to present the merge request review results, if not, use Markdown formatting.
    return prompt


def get_feature_development_prompt(
    jira_project_key: str,
    feature_description: str,
) -> str:
    """A prompt for feature development workflow setup."""
    feature_description = wrap_with_xml_tag("feature_description", feature_description)
    prompt = inspect.cleandoc(f"""
        Help a user setup a new feature branch by following these steps:

        ### Create a Jira issue for the feature in the project {jira_project_key}
        - Use the provided feature description: {feature_description} to prepare a summary and description for a new Jira issue, 
        ask for additional details if needed. 
        - Use tool to fetch allowed issue types in the project. Based on description and allowed types, suggest the most 
        appropriate types for the issue.
        - Try using local context to determine the user's name in Jira by inspecting local Git repo metadata via tools,
        if available, look at user emails in the Git config. If unsuccessful, ask for the user's Jira username directly. 
        Use the emails or provided username to search for the user in Jira to find or validate the username.
        - Provide the user with arguments to the Jira issue creation tool, after approval create a new Jira issue and
        provide the user with the created issue key and URL.
        
        ### Create a new feature branch in GitLab
        - Try inferring the GitLab project from local Git context by inspecting the remote urls from the get repo metadata tool,
        if unsuccessful, ask for the GitLab project path or url directly.
        - Build the feature branch name using the created Jira issue key: "feature/{{jira_issue_key}}-{{issue_summary}}".
        - The source branch in most cases should be "testing", but check that it exists in the project using the get project info tool.
        - Provide the user with arguments to the branch creation tool, after approval create the branch.

        ### Create a merge request in GitLab
        - Use the GitLab project path, source and target branch from the previous steps.
        - Use the created Jira issue description to create merge request title and description, the title should start with the 
        issue key.
        - Use the Jira username to find the user id in GitLab via the search users tool.
        - The merge request description should start with line "Closes {{jira_issue_key}}".
        - Provide the user with arguments to the merge request creation tool, after approval create the merge request and 
        provide the user with the created merge request url.
        - Use the {GitLabTools.search_merge_requests.__name__} tool to check whether an open merge request from testing to master
        branch exists, if not, suggest to create one.

        ### Add backward compatibility tests in GitLab
        - Ask user if he wants to add backward compatibility or other tests to the feature branch, either in GitLab or in local 
        development environment (only if local tools for editing files are available). If user doesn't agree or prefers local 
        development environment, skip this step.
        - Use tools to inspect code in the newly created GitLab feature branch.
        - Use create GitLab file tool to add tests to the feature branch.

        ### Pull the feature branch
        - If there are not local Git tools available, skip pulling the branch.
        - Inspect the local Git repo metadata and see if there is a remote that points to the GitLab project in which the 
        feature branch was created. If not, skip pulling the branch.
        - Suggest to user to pull the newly created feature branch, if approved, use the pull branch tool with the 
        feature branch name from the previous step.

        ### Add tests and initial changes locally
        - If local file editing tools are not available or user doesn't want to add tests and initial changes, skip this step.
        - Inspect local codebase. Based on feature description, determine which parts of the code are likely to change.
        - Based on the codebase and likely changes, add backward compatibilty tests to ensure that the feature doesn't break 
        existing functionality.
        - Based on the codebase and likely changes, add initial development changes.
    """)
    return prompt



import datetime
import functools
import inspect
import json
import logging
import os
import re
import socket
import ssl
import tempfile
import textwrap
import typing
import urllib.parse
import warnings
import zoneinfo
from typing import Callable, Iterable

import fastmcp.tools
import httpx
import langchain.tools
import langchain_core.utils.function_calling
import mcp
import pydantic
import sqlalchemy as sa
import urllib3


def truncate_text(text: str, max_length: int, placeholder: str = "[...]") -> str:
    text = text[:max_length] + placeholder if len(text) > max_length else text
    return text


def json_dumps_for_ai(
    value: str | list | dict | pydantic.BaseModel,
    max_length: int | None = None,
    truncation_placeholder: str = "\n[CONTENT END] The rest of the content was truncated to fit into LLM context window.",
    **kwargs,
) -> str:
    if isinstance(value, str):
        return value

    if isinstance(value, pydantic.BaseModel):
        value = value.model_dump()

    value = json.dumps(value, ensure_ascii=False, **kwargs)
    if max_length:
        value = truncate_text(value, max_length, placeholder=truncation_placeholder)

    return value


def format_datetime_for_ai(
    date: datetime.datetime | str,
    timespec: str = "seconds",
    timezone: zoneinfo.ZoneInfo = zoneinfo.ZoneInfo("UTC"),
) -> str:
    date = datetime.datetime.fromisoformat(date) if isinstance(date, str) else date
    date = date.astimezone(timezone)
    date = date.isoformat(timespec=timespec, sep=" ")
    return date


def to_formatted_markdown(text: str, kind: str = "json") -> str:
    text = f"```{kind}\n{text}\n```"
    return text


def format_json_for_ai(
    value: str | list | dict | pydantic.BaseModel, indent: int | None = 0
) -> str:
    text = json_dumps_for_ai(value, indent=indent)
    text = to_formatted_markdown(text)
    return text


def filter_dict_by_keys(dictionary: dict, keys: Iterable) -> dict:
    dictionary = {key: dictionary[key] for key in keys if key in dictionary}
    return dictionary


def recursively_collect_file_paths_in_directory(directory_path: str) -> list[str]:
    file_paths = []
    for root, _, files in os.walk(directory_path):
        for file in files:
            full_path = os.path.join(root, file)
            if os.path.isfile(full_path):
                file_paths.append(full_path)

    return file_paths


def get_return_type_annotation(function):
    return function.__annotations__.get("return")


def isinstance_typing_annotation(annotation, type) -> bool:
    return annotation is type or typing.get_origin(annotation) is type


def to_fastmcp_tool(
    function,
    tags: set[str] | None,
    annotations: mcp.types.ToolAnnotations | None,
    max_output_length: int | None = None,
) -> fastmcp.tools.FunctionTool:
    """Adds a custom serializer to ensure consistent and token-efficient conversion of tool output to text."""
    return_type = get_return_type_annotation(function)
    if isinstance_typing_annotation(return_type, list):
        warnings.warn(
            "When tool returns a list, each of its elements will be serialized separately "
            "using fastmcp.tools.tool._convert_to_content into MCPContent."
        )

    tool = fastmcp.tools.FunctionTool.from_function(
        function,
        tags=tags,
        annotations=annotations,
        serializer=functools.partial(json_dumps_for_ai, max_length=max_output_length),
    )
    # Normalize tool parameters schema to match OpenAI's.
    # For example, it dereferences JSON schema $refs, which are not supported by many AI API providers.
    openai_tool_schema = langchain_core.utils.function_calling.convert_to_openai_tool(function)
    tool.parameters = openai_tool_schema["function"]["parameters"]
    return tool


def request_as_dict(*args, **kwargs) -> dict:
    response = httpx.get(*args, **kwargs)
    response.raise_for_status()
    response_dict = response.json()
    return response_dict


def to_datetime(
    date: datetime.date | datetime.timedelta | str | int | float | None = None,
) -> datetime.datetime:
    match date:
        case datetime.datetime():
            pass
        case datetime.date():
            date = datetime.datetime.combine(date, datetime.time())
        case None:
            date = datetime.datetime.now()
        case datetime.timedelta():
            date += datetime.datetime.now()
        case str():
            date = datetime.datetime.fromisoformat(date)
        case int() | float():
            date = datetime.datetime.fromtimestamp(date)
        case _:
            raise ValueError(f"Unexpected date type: {type(date)}")

    date = date if date.tzinfo else date.astimezone()
    # or date = date.astimezone(datetime.timezone.utc)
    return date


def join_url_components(*url_components: str) -> str:
    url = "/".join(str(i).strip("/") for i in url_components)
    return url


def try_format_json_with_indent(text: str, indent: int = 2) -> str:
    try:
        text = json.dumps(json.loads(text), indent=indent)
    except json.JSONDecodeError:
        pass

    return text


def raise_for_status(response: httpx.Response, max_text_length: int = 1000) -> None:
    if not response.is_error:
        return

    response_text = try_format_json_with_indent(response.text)
    request_text = try_format_json_with_indent(response.request.content.decode())
    if max_text_length:
        request_text = truncate_text(request_text, max_text_length)
        response_text = truncate_text(response_text, max_text_length)

    error_message = (
        f"Error in HTTP response while requesting {response.url}\n"
        f"Status code: {response.status_code}\n"
        f"Reason: {response.reason_phrase}\n"
        f"Request body: {request_text}\n"
        f"Response body: {response_text}\n"
    )
    error = httpx.HTTPStatusError(error_message, request=response.request, response=response)
    raise error


async def araise_for_status(response: httpx.Response, max_response_text_length: int = 1000) -> None:
    if not response.is_error:
        return

    await response.aread()
    raise_for_status(response, max_text_length=max_response_text_length)


def try_match_regex(pattern: str, string: str, **kwargs):
    match = re.match(pattern, string, **kwargs)
    if not match:
        raise ValueError(f"String '{string}' does not fit expected pattern '{pattern}'")

    return match


def value_to_key(items: list, key: str) -> dict:
    dictionary = {i.pop(key): i for i in items}
    return dictionary


def to_maybe_ssl_context(ssl_verify: bool | str) -> bool | ssl.SSLContext:
    if isinstance(ssl_verify, str):
        ssl_verify = ssl.create_default_context(cafile=ssl_verify)

    return ssl_verify


def add_mcp_server_components(
    server: fastmcp.FastMCP | None = None,
    tools: Iterable[fastmcp.tools.FunctionTool] = (),
    resources: Iterable[fastmcp.resources.Resource] = (),
    prompts: Iterable[fastmcp.prompts.Prompt] = (),
) -> fastmcp.FastMCP:
    server = server or fastmcp.FastMCP()
    for tool in tools:
        server.add_tool(tool)

    for resource in resources:
        server.add_resource(resource)

    for prompt in prompts:
        server.add_prompt(prompt)

    return server


def wrap_with_xml_tag(tag: str, value: str, with_new_lines: bool = False) -> str:
    value = f"\n{value}\n" if with_new_lines else value
    wrapped_value = f"<{tag}>{value}</{tag}>"
    wrapped_value += "\n" if with_new_lines else ""
    return wrapped_value


def assign_default_tool_annotation_hints(
    annotation: mcp.types.ToolAnnotations | None = None,
) -> mcp.types.ToolAnnotations:
    annotation = annotation or mcp.types.ToolAnnotations()
    if annotation.readOnlyHint and annotation.destructiveHint:
        raise ValueError("Tool cannot be both readonly and destructive.")

    if annotation.readOnlyHint and annotation.idempotentHint is False:
        raise ValueError("Tool cannot be both readonly and non-idempotent.")

    default = mcp.types.ToolAnnotations(
        readOnlyHint=False,
        destructiveHint=True,
        idempotentHint=False,
        openWorldHint=True,
    )
    for hint, value in default.model_dump().items():
        if getattr(annotation, hint) is None:
            setattr(annotation, hint, value)

    return annotation


def disable_urllib_insecure_request_warning():
    urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


def set_ssl_cert_file_from_cadata(cadata: str, set_ssl_cert_file_env_var: bool = True) -> str:
    with tempfile.NamedTemporaryFile(delete=False, mode="w", suffix=".pem") as temp_ca_file:
        temp_ca_file.write(cadata)
        ca_file_path = temp_ca_file.name

    if set_ssl_cert_file_env_var:
        os.environ["SSL_CERT_FILE"] = ca_file_path

    return ca_file_path


def split_schema_table(schema_table: str) -> tuple[str | None, str]:
    schema, table = schema_table.split(".") if "." in schema_table else (None, schema_table)
    return schema, table


def get_all_schema_tables(engine: sa.Engine, verify_select_permissions: bool = False) -> list[str]:
    inspector = sa.inspect(engine)
    schemas = inspector.get_schema_names()
    schema_tables = []
    with engine.connect() as connection:
        for schema in schemas:
            try:
                tables = inspector.get_table_names(schema)
            except sa.exc.DatabaseError as e:
                logging.warning(f"Failed to get table names for schema {schema}:\n{e}")
                continue

            if not verify_select_permissions:
                schema_tables.extend([f"{schema}.{table}" for table in tables])
                continue

            for table in tables:
                # Use text to avoid expensive reflection for each table.
                probe_sql = sa.text(f'SELECT 1 FROM "{schema}"."{table}" LIMIT 0')
                try:
                    connection.execute(probe_sql)
                except sa.exc.DatabaseError:
                    # Need to rollback failed connection state to continue executing commands.
                    connection.rollback()
                else:
                    schema_tables.append(f"{schema}.{table}")

    return schema_tables


class StringFilter(pydantic.BaseModel):
    allowed_values: set[str] = set()
    forbidden_values: set[str] = set()
    allowed_pattern: re.Pattern[str] | None = None
    forbidden_pattern: re.Pattern[str] | None = None

    @classmethod
    def from_lists_and_strings(
        cls,
        allowed_values: list[str] | None = None,
        forbidden_values: list[str] | None = None,
        allowed_pattern: str | None = None,
        forbidden_pattern: str | None = None,
    ):
        string_filter = cls(
            allowed_values=set(allowed_values or []),
            forbidden_values=set(forbidden_values or []),
            allowed_pattern=re.compile(allowed_pattern) if allowed_pattern else None,
            forbidden_pattern=re.compile(forbidden_pattern) if forbidden_pattern else None,
        )
        return string_filter

    def is_explicitly_allowed(self, value: str) -> bool:
        return (
            value in self.allowed_values
            or self.allowed_pattern
            and self.allowed_pattern.fullmatch(value)
        )

    def is_explicitly_forbidden(self, value: str) -> bool:
        return (
            value in self.forbidden_values
            or self.forbidden_pattern
            and self.forbidden_pattern.fullmatch(value)
        )

    def should_keep(self, value: str) -> bool:
        return self.is_explicitly_allowed(value) and not self.is_explicitly_forbidden(value)


def find_available_port(host: str = "127.0.0.1") -> int:
    """Return an OS-allocated *free* TCP port number for *host*.

    This helper opens a temporary socket bound to *(host, 0)*: passing
    ``0`` tells the kernel to choose any port that is currently unused.
    After binding we read the port number via ``sock.getsockname()`` and
    immediately close the socket by leaving the *with* block.  The port is
    therefore released back to the OS, ready for the caller to reuse.

    Parameters
    ----------
    host : str, default "127.0.0.1"
        Interface to bind to.  Normally ``"127.0.0.1"`` (IPv4 loopback) or
        ``"::1"`` (IPv6 loopback).  Using an explicit host avoids the risk of
        exposing the service on a public interface.

    Returns
    -------
    int
        A currently free TCP port number.

    Notes
    -----
    ``socket.AF_INET`` specifies the IPv4 address family, while
    ``socket.SOCK_STREAM`` selects the TCP protocol.  Together they create a
    TCP/IPv4 socket.

    This pattern is perfectly adequate for local development, but in highly
    concurrent environments there is still a small race window between
    releasing the port and rebinding to it.  If you need stronger guarantees
    consider opening the socket once and *passing the open file descriptor*
    (e.g., via ``sock.fileno()``) to the server process instead of closing it
    here.
    """
    with socket.socket(family=socket.AF_INET, type=socket.SOCK_STREAM) as sock:
        sock.bind((host, 0))
        port = sock.getsockname()[1]

    return port


async def llm_friendly_http_request(
    client: httpx.AsyncClient,
    url: str,
    http_method: str = "GET",
    json_query_params: str | None = None,
    content: str | dict | None = None,
    ensure_dict_output: bool = False,
):
    json_query_params = json.loads(json_query_params) if json_query_params else None
    kwargs = dict(
        method=http_method,
        url=url,
        params=json_query_params,
    )
    match content:
        case str():
            try:
                kwargs["json"] = json.loads(content)
            except json.JSONDecodeError:
                kwargs["content"] = content
        case dict():
            kwargs["json"] = content

    response = await client.request(**kwargs)
    data = response.json()
    data = {"response": data} if ensure_dict_output and not isinstance(data, dict) else data
    return data


def prepend_url_path_prefix_if_not_present(prefix: str, path: str) -> str:
    path = path.strip("/")
    prefix = prefix.strip("/")
    path = path if path.startswith(prefix) else f"{prefix}/{path}"
    return path


def wrap_tool_with_str_serializer(
    tool: callable, max_output_length: int | None = None, **kwargs
) -> callable:
    def serialize_output(output) -> str:
        return json_dumps_for_ai(output, max_length=max_output_length, **kwargs)

    @functools.wraps(tool)
    async def awrapped(*args, **tool_kwargs):
        output = await tool(*args, **tool_kwargs)
        output = serialize_output(output)
        return output

    @functools.wraps(tool)
    def wrapped(*args, **tool_kwargs):
        output = tool(*args, **tool_kwargs)
        output = serialize_output(output)
        return output

    wrapped_tool = awrapped if inspect.iscoroutinefunction(tool) else wrapped
    wrapped_tool.__annotations__["return"] = str
    return wrapped_tool


def get_line_separated_text_overview(
    text: str,
    line_chunks_len: int = 5,
    separator: str = "[LINES TRUNCATED FOR BREVITY]",
) -> str:
    """Return head, tail and middle of the text"""
    text = textwrap.dedent(text)
    lines = [line.rstrip() for line in text.splitlines() if line.rstrip()]
    if len(lines) < (line_chunks_len * 3) + 10:
        text = "\n".join(lines)
        return text

    head = lines[:line_chunks_len]
    mid = lines[(len(lines) - line_chunks_len) // 2 : (len(lines) + line_chunks_len) // 2]
    tail = lines[-line_chunks_len:]

    overview = "\n".join(head + [separator] + mid + [separator] + tail)
    return overview


def to_langchain_tool(
    tool: Callable | fastmcp.tools.FunctionTool,
    prefix: str | None = None,
    max_output_length: int | None = None,
) -> langchain.tools.BaseTool:
    match tool:
        case fastmcp.tools.FunctionTool(fn=fn, annotations=annotations, tags=tags):
            tool = fn
            metadata = annotations.model_dump(exclude_none=True) if annotations else None
            tags = sorted(tags) if tags else None
        case _:
            metadata = None
            tags = None

    if isinstance_typing_annotation(get_return_type_annotation(tool), dict):
        # Serialize dict outputs to JSON.
        tool = wrap_tool_with_str_serializer(tool, max_output_length=max_output_length)

    tool = langchain.tools.tool(tool)
    tool.name = f"{prefix}_{tool.name}" if prefix else tool.name
    tool.tags = tags
    tool.metadata = metadata
    return tool


def build_url(*url_components: str, **kwargs: str) -> str:
    url = join_url_components(*url_components)
    url = f"{url}?{urllib.parse.urlencode(kwargs)}" if kwargs else url
    return url



import datetime


def to_grafana_time_format(date: str) -> str | int:
    try:
        date = datetime.datetime.fromisoformat(date)
    except ValueError:
        pass
    else:
        date = int(date.timestamp() * 1000)

    return date



import inspect

from bir_mcp.grafana.resources import get_flux_instruction, get_mcp_resource_uri_functions


def get_prompts() -> list[callable]:
    prompts = [
        get_lowest_network_usage_virtual_machines_prompt,
    ]
    return prompts


def get_lowest_network_usage_virtual_machines_prompt() -> str:
    """The prompt for finding virtual machines with the lowest network usage."""
    uri_functions = get_mcp_resource_uri_functions()
    flux_instruction_uri = uri_functions.inverse[get_flux_instruction]
    prompt = inspect.cleandoc(f"""
        Which virtual machines had the lowest network usage over the last week in the bank, 
        based on Vsphere logs in InfluxDB? To answer this question, follow these steps:
        - Refer to Flux language instructions in the MCP resources {flux_instruction_uri}.
        - Find InfluxDB datasource in the Grafana instance.
        - Find the InfluxDB datasource, bucket and measurement name that relates to Vsphere virtual machines usage in the bank.
        - Find the tag that identifies virtual machines.
        - Construct a query for finding the virtual machines with the lowest network usage using the lowestAverage() Flux function.
        - Present the results.
    """)
    return prompt



import enum
import functools
import inspect
import warnings
from typing import Annotated

import fastmcp
import typer

import bir_mcp.atlassian
import bir_mcp.charts
import bir_mcp.git_lab
import bir_mcp.grafana
import bir_mcp.sql
from bir_mcp.config import SystemManagedConfig


class McpCategory(enum.StrEnum):
    gitlab = enum.auto()
    grafana = enum.auto()
    jira = enum.auto()
    local = enum.auto()


def build_mcp_server(
    enable_local_tools: Annotated[
        bool,
        typer.Option(
            help=inspect.cleandoc(
                """
                Whether to include local tools, such as those that interact with local
                file system and perform local Git actions.
                """
            ),
        ),
    ] = False,
    include_non_readonly_tools: Annotated[
        bool,
        typer.Option(
            help=inspect.cleandoc(
                """
                Whether to include tools with ability to write and create new resources,
                for example new Jira issues, GitLab branches or commits.
                """
            ),
        ),
    ] = False,
    include_destructive_tools: Annotated[
        bool,
        typer.Option(
            help=inspect.cleandoc(
                """
                Whether to include tools with ability to delete or overwrite resources,
                such as GitLab file updates.
                """
            ),
        ),
    ] = False,
    consul_host: Annotated[
        str | None,
        typer.Option(
            help=inspect.cleandoc(
                """
                The host of the Consul key-value store to load the system managed configuration from,
                with the SQL connections and default system connection credentials.
                Any parameters explicity provided to this function will override the ones
                loaded from Consul.
                """
            ),
        ),
    ] = None,
    consul_key: Annotated[
        str | None,
        typer.Option(help="The key name of the Consul key-value store."),
    ] = None,
    consul_token: Annotated[
        str | None,
        typer.Option(help="The token to use for authentication with Consul."),
    ] = None,
    gitlab_private_token: Annotated[
        str | None,
        typer.Option(
            help=inspect.cleandoc(
                """
                The GitLab token to use for authentication with GitLab. To enable writing tools,
                the token should have write API permissions.
                """
            ),
        ),
    ] = None,
    grafana_username: str | None = None,
    grafana_password: str | None = None,
    grafana_token: Annotated[
        str | None,
        typer.Option(
            help=inspect.cleandoc(
                """
                The Grafana API token to use for authentication with Grafana.
                If provided, this will be used instead of username/password authentication.
                """
            ),
        ),
    ] = None,
    jira_token: str | None = None,
    gitlab_url: str = "https://gitlab.kapitalbank.az",
    grafana_url: str = "https://yuno.kapitalbank.az",
    jira_url: str = "https://jira-support.kapitalbank.az",
    timezone: Annotated[
        str,
        typer.Option(
            help=inspect.cleandoc(
                """
                The timezone to use for datetime formatting in the output of tools.
                """
            ),
        ),
    ] = "Asia/Baku",
    tools_max_output_length: Annotated[
        int | None,
        typer.Option(
            help=inspect.cleandoc(
                """
                The maximum output string length for tools, with longer strings being truncated.
                Especially useful for AI clients with smaller context windows and those that
                cannot automatically trim or summarize chat history, for example Claude Desktop.
                """
            )
        ),
    ] = None,
    verify_ssl: Annotated[bool, typer.Option(help="Whether to verify SSL certificates.")] = True,
    ca_file: Annotated[
        str | None,
        typer.Option(help="Path to a local certificate authority file to verify SSL certificates."),
    ] = None,
    prompts_to_tools: Annotated[
        bool,
        typer.Option(
            help=inspect.cleandoc(
                """
                Whether to convert prompts to tools. Not all MCP clients support prompts (for example
                Windsurf at the time of writing), but all of them support tools.
                """
            )
        ),
    ] = False,
) -> fastmcp.FastMCP:
    ssl_verify = ca_file or verify_ssl  # Workaround because typer doesn't support union types.
    if consul_host and consul_key:
        config = SystemManagedConfig.from_consul(
            consul_host=consul_host,
            consul_token=consul_token,
            consul_key=consul_key,
        )
    else:
        config = SystemManagedConfig()

    config_overrides = dict(
        gitlab_private_token=gitlab_private_token,
        grafana_token=grafana_token,
        grafana_username=grafana_username,
        grafana_password=grafana_password,
        jira_token=jira_token,
        gitlab_url=gitlab_url,
        grafana_url=grafana_url,
        jira_url=jira_url,
    )
    config_overrides = {k: v for k, v in config_overrides.items() if v}
    config = config.model_copy(update=config_overrides)

    server = fastmcp.FastMCP(
        name="Bir MCP server",
        instructions=inspect.cleandoc("""
            MCP server for BirBank.
        """),
    )
    for sub_mcp in generate_mcps(
        config=config,
        timezone=timezone,
        enable_local_tools=enable_local_tools,
        ssl_verify=ssl_verify,
    ):
        subserver = sub_mcp.get_mcp_server(
            include_non_readonly_tools=include_non_readonly_tools,
            include_destructive_tools=include_destructive_tools,
            max_output_length=tools_max_output_length,
            prompts_to_tools=prompts_to_tools,
        )
        server.mount(subserver, prefix=sub_mcp.get_tag())

    return server


def generate_mcps(
    config: SystemManagedConfig,
    timezone: str,
    enable_local_tools: bool,
    ssl_verify: bool | str,
):
    common_params = dict(timezone=timezone)
    if enable_local_tools:
        from bir_mcp.local.filesystem import Filesystem
        from bir_mcp.local.git_local import Git

        for local_mcp in [
            Git(**common_params),
            Filesystem(**common_params),
        ]:
            yield local_mcp

    for sql_context in config.sql_contexts:
        yield bir_mcp.sql.SQL(sql_context=sql_context, **common_params)

    yield bir_mcp.charts.Charts(
        engines={name: conn.get_engine() for name, conn in config.sql_connections.items()},
        **common_params,
    )

    common_params["ssl_verify"] = ssl_verify
    if config.gitlab_private_token:
        yield bir_mcp.git_lab.GitLab(
            url=config.gitlab_url, private_token=config.gitlab_private_token, **common_params
        )
    else:
        warnings.warn(
            "Since GitLab private token is not provided, the GitLab tools will not be available."
        )

    if grafana_token_or_auth := config.get_grafana_token_or_auth():
        yield bir_mcp.grafana.Grafana(
            url=config.grafana_url,
            token_or_auth=grafana_token_or_auth,
            **common_params,
        )
    else:
        warnings.warn(
            "Since Grafana token or username/password are not provided, the Grafana tools will not be available."
        )

    if config.jira_token:
        yield bir_mcp.atlassian.Atlassian(token=config.jira_token, **common_params)
    else:
        warnings.warn(
            "Since Jira token is not provided, the Atlassian tools will not be available."
        )

    if config.sonarqube_token:
        yield bir_mcp.git_lab.SonarQube(token=config.sonarqube_token, **common_params)
    else:
        warnings.warn(
            "Since SonarQube token is not provided, the SonarQube tools will not be available."
        )


@functools.wraps(build_mcp_server)
def build_and_run(*args, **kwargs):
    server = build_mcp_server(*args, **kwargs)
    server.run()


def main():
    typer.run(build_and_run)


if __name__ == "__main__":
    main()



import abc
import datetime
import inspect
import zoneinfo
from typing import Annotated, Iterable

import fastmcp.prompts
import fastmcp.resources
import fastmcp.tools
import httpx
import langchain.tools
import mcp
import pydantic

from bir_mcp.utils import (
    add_mcp_server_components,
    assign_default_tool_annotation_hints,
    format_datetime_for_ai,
    llm_friendly_http_request,
    prepend_url_path_prefix_if_not_present,
    to_fastmcp_tool,
    to_langchain_tool,
)


def build_tools(
    functions: Iterable[callable],
    annotations: mcp.types.ToolAnnotations | None = None,
    max_output_length: int | None = None,
    tags: set[str] | None = None,
) -> list[fastmcp.tools.FunctionTool]:
    tools = [
        to_fastmcp_tool(
            function,
            tags=tags,
            annotations=annotations,
            max_output_length=max_output_length,
        )
        for function in functions
    ]
    return tools


def build_readonly_tools(
    functions: Iterable[callable],
    max_output_length: int | None = None,
    tags: set[str] | None = None,
) -> list[fastmcp.tools.FunctionTool]:
    tools = build_tools(
        functions=functions,
        annotations=mcp.types.ToolAnnotations(
            readOnlyHint=True, destructiveHint=False, idempotentHint=True
        ),
        max_output_length=max_output_length,
        tags=tags,
    )
    return tools


def build_write_tools(
    functions: Iterable[callable],
    max_output_length: int | None = None,
    tags: set[str] | None = None,
) -> list[fastmcp.tools.FunctionTool]:
    tools = build_tools(
        functions=functions,
        annotations=mcp.types.ToolAnnotations(
            readOnlyHint=False, destructiveHint=False, idempotentHint=False
        ),
        max_output_length=max_output_length,
        tags=tags,
    )
    return tools


def build_resources(
    resource_uri_functions: dict[str, callable], tags: set[str] | None = None
) -> list[fastmcp.resources.Resource]:
    resources = [
        fastmcp.resources.Resource.from_function(function, uri=uri, tags=tags)
        for uri, function in resource_uri_functions.items()
    ]
    return resources


def build_prompts(
    prompts: Iterable[callable], tags: set[str] | None = None
) -> list[fastmcp.prompts.Prompt]:
    prompts = [fastmcp.prompts.Prompt.from_function(function, tags=tags) for function in prompts]
    return prompts


class BaseMcp(abc.ABC):
    def __init__(self, timezone: str = "UTC"):
        self.timezone = zoneinfo.ZoneInfo(timezone)

    def format_datetime_for_ai(self, date: datetime.datetime | str) -> str:
        return format_datetime_for_ai(date, timezone=self.timezone)

    def get_tag(self) -> str | None:
        return

    def get_mcp_server_without_components(self) -> fastmcp.FastMCP:
        return fastmcp.FastMCP()

    def get_tools(self) -> list[callable]:
        return []

    def get_uri_resources(self) -> dict[str, callable]:
        return {}

    def get_prompts(self) -> list[callable]:
        return []

    def get_mcp_tools(
        self, max_output_length: int | None = None, tags: set[str] | None = None
    ) -> list[fastmcp.tools.FunctionTool]:
        tools = self.get_tools()
        tools = build_tools(functions=tools, max_output_length=max_output_length, tags=tags)
        return tools

    def get_mcp_resources(self, tags: set[str] | None = None) -> list[fastmcp.resources.Resource]:
        uri_resources = self.get_uri_resources()
        resources = build_resources(resource_uri_functions=uri_resources, tags=tags)
        return resources

    def get_mcp_prompts(self, tags: set[str] | None = None) -> list[fastmcp.prompts.Prompt]:
        prompts = self.get_prompts()
        prompts = build_prompts(prompts=prompts, tags=tags)
        return prompts

    def get_mcp_server(
        self,
        server_to_extend: fastmcp.FastMCP | None = None,
        include_non_readonly_tools: bool = False,
        include_destructive_tools: bool = False,
        max_output_length: int | None = None,
        tags: set[str] | None = None,
        prompts_to_tools: bool = False,
    ) -> fastmcp.FastMCP:
        server = server_to_extend or self.get_mcp_server_without_components()
        tags = {self.get_tag()} if tags is None and self.get_tag() else tags
        tools = self.get_mcp_tools(max_output_length=max_output_length, tags=tags)
        filtered_tools = []
        for tool in tools:
            tool.annotations = assign_default_tool_annotation_hints(tool.annotations)
            if not include_non_readonly_tools and not tool.annotations.readOnlyHint:
                continue
            if not include_destructive_tools and tool.annotations.destructiveHint:
                continue
            filtered_tools.append(tool)

        resources = self.get_mcp_resources(tags=tags)
        prompts = self.get_mcp_prompts(tags=tags)
        if prompts_to_tools:
            filtered_tools.extend(build_readonly_tools([p.fn for p in prompts], tags=tags))

        server = add_mcp_server_components(
            server=server,
            tools=filtered_tools,
            resources=resources,
            prompts=prompts,
        )
        return server

    def get_langchain_tools(
        self,
        max_output_length: int | None = None,
        with_prefix: bool = True,
    ) -> list[langchain.tools.BaseTool]:
        tools = []
        for tool in self.get_tools():
            tool = to_langchain_tool(
                tool,
                prefix=self.get_tag() if with_prefix else None,
                max_output_length=max_output_length,
            )
            tools.append(tool)

        return tools


class AhttpxMixin:
    def prepend_api(self, path: str) -> str:
        return prepend_url_path_prefix_if_not_present("api", path)

    async def ahttpx_get(self, endpoint: str, **kwargs) -> httpx.Response:
        endpoint = self.prepend_api(endpoint)
        response = await self.ahttpx.get(url=endpoint, params=kwargs)
        return response

    async def ahttpx_post(self, endpoint: str, json: dict) -> httpx.Response:
        endpoint = self.prepend_api(endpoint)
        response = await self.ahttpx.post(url=endpoint, json=json)
        return response

    async def http_request(
        self,
        path: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The path to the API endpoint, relative to the base URL.
                    Should not include the api route prefix.
                """)
            ),
        ],
        http_method: str = "GET",
        json_query_params: Annotated[
            str | None,
            pydantic.Field(description="The url query parameters dict, serialized to JSON."),
        ] = None,
        content: Annotated[
            str | dict | None,
            pydantic.Field(description="The request body."),
        ] = None,
    ) -> dict:
        """
        Performs a HTTP request to the REST API. Can be used when no specialized tool is available for the task.
        The arguments are passed to a httpx.Client.request method, with base_url set to the local
        instance's url and API key set in headers, so the result looks like this:
        return client.request(method=http_method, url=path, params=json.loads(json_query_params), content=content).json()
        """
        data = await llm_friendly_http_request(
            client=self.ahttpx,
            url=self.prepend_api(path),
            http_method=http_method,
            json_query_params=json_query_params,
            content=content,
            ensure_dict_output=True,
        )
        return data



import base64
import inspect
import io
import pathlib
import tempfile
import threading
import uuid
from typing import override

import altair
import fastapi.staticfiles
import fastmcp
import mcp
import pandas as pd
import sqlalchemy as sa
import uvicorn

from bir_mcp.core import BaseMcp
from bir_mcp.utils import find_available_port


class Charts(BaseMcp):
    def __init__(
        self,
        engines: dict[str, sa.Engine],
        timezone: str = "UTC",
        max_served_static_files: int = 100,
    ):
        super().__init__(timezone=timezone)
        self.engines = engines

        # Initialize a background FastAPI + Uvicorn server to serve generated HTML chart files
        self._max_served_files = max_served_static_files
        self._serve_dir_obj = tempfile.TemporaryDirectory()
        self.serve_dir = pathlib.Path(self._serve_dir_obj.name)
        self.port = find_available_port()
        self.app = fastapi.FastAPI()
        self.app.mount(
            "/",
            fastapi.staticfiles.StaticFiles(directory=str(self.serve_dir), html=True),
            name="static",
        )

        def _run() -> None:
            uvicorn.run(
                self.app,
                host="127.0.0.1",
                port=self.port,
                log_level="warning",
                lifespan="off",  # faster startup; we don't need lifespan events
            )

        self.thread = threading.Thread(target=_run, daemon=True)
        self.thread.start()

    @override
    def get_tag(self):
        return "charts"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="MCP server with charting and plotting tools",
            instructions=inspect.cleandoc("""
                Contains tools to create charts and plots.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        tools = [
            self.build_chart,
            self.build_html_page,
        ]
        tools = [
            fastmcp.tools.FunctionTool.from_function(
                tool,
                tags=tags,
                annotations=mcp.types.ToolAnnotations(
                    readOnlyHint=True, destructiveHint=False, idempotentHint=True
                ),
            )
            for tool in tools
        ]
        return tools

    def _cleanup_old_files(self) -> None:
        paths = self.serve_dir.iterdir()
        paths = sorted(paths, key=lambda path: path.stat().st_mtime)
        for path in paths[-self._max_served_files :]:
            path.unlink(missing_ok=True)

    def _build_link(self, path: str, name: str | None = None) -> str:
        link = f"http://127.0.0.1:{self.port}/{path}"
        link = f"[{name}]({link})" if name else link
        return link

    def build_chart(self, vega_lite_json_spec: str, sql_query: str, connection_name: str):
        """
        Build a single chart from a Vega-Lite JSON specification and a SQL query.
        The connection name can be determined by calling the get_database_info tool of an SQL MCP.
        Refer to data source as "sql_query", for example:
        ```json
        {
        "$schema": "https://vega.github.io/schema/vega-lite/v5.json",
        "data": {"name": "sql_query"},
        "title": "My Bar Chart",
        "mark": "bar",
        "encoding": {
            "x": {"field": "category", "type": "nominal", "title": "Category"},
            "y": {"field": "value", "type": "quantitative", "title": "Value"},
            "color": {"field": "group", "type": "nominal", "title": "Group"}
        },
        "width": 400,
        "height": 300
        }
        ```
        Spaces in the example JSON are just for human readability, when generating the real JSON,
        omit all spaces.
        """
        # Json schema https://vega.github.io/schema/vega-lite/v5.json is huge, need RAG for it.
        # After generating PNG chart may need to filter it from chat conversation to save tokens
        if not (engine := self.engines.get(connection_name)):
            raise ValueError(
                f"Connection {connection_name} not found among available connections: {sorted(self.engines)}"
            )

        df = pd.read_sql(sql_query, engine)
        chart = altair.Chart.from_json(vega_lite_json_spec)
        chart.data = df
        io_bytes = io.BytesIO()
        chart.save(io_bytes, format="png")
        image = mcp.types.ImageContent(
            type="image",
            mimeType="image/png",
            data=base64.b64encode(io_bytes.getvalue()).decode(),
        )

        self._cleanup_old_files()
        filename = f"chart_{uuid.uuid4().hex}.html"
        file_path = self.serve_dir / filename
        chart.save(file_path, format="html")
        text = mcp.types.TextContent(
            type="text",
            text=self._build_link(name="Chart", path=filename),
        )

        return [image, text]

    def build_html_page(self, html: str):
        """Build and serve an HTML page. Returns a link to the page."""
        self._cleanup_old_files()
        filename = f"page_{uuid.uuid4().hex}.html"
        file_path = self.serve_dir / filename
        # TODO: switch to anyio.read_file for async, but it's probably a minor optimization.
        with open(file_path, "w") as f:
            f.write(html)

        text = mcp.types.TextContent(
            type="text",
            text=self._build_link(name="Page", path=filename),
        )
        return text



import os

import consul
import pydantic

from bir_mcp.utils import disable_urllib_insecure_request_warning


class ConsulKeyValue(pydantic.BaseModel):
    host: str
    key: str
    token: str | None = None
    scheme: str = "https"
    port: int = 443
    verify: bool = False  # Our Consul instances don't have valid SSL certificates.

    @classmethod
    def from_env_variables(cls):
        host = os.environ["CONSUL_HOST"]
        token = os.environ.get("CONSUL_TOKEN")
        key = os.environ["CONSUL_KEY"]
        consul_key_value = cls(host=host, token=token, key=key)
        return consul_key_value

    def model_post_init(self, _context) -> None:
        if not self.verify:
            disable_urllib_insecure_request_warning()

    def get_client(self) -> consul.Consul:
        client = consul.Consul(
            host=self.host,
            port=self.port,
            token=self.token,
            scheme=self.scheme,
            verify=self.verify,
        )
        return client

    def load(self) -> bytes:
        client = self.get_client()
        index, data = client.kv.get(self.key)
        value = data["Value"]
        return value



import re


def infer_jira_field_representation(
    value, dict_key_preference=("displayName", "name", "value", "key")
):
    def recursive_infer(value):
        match value:
            case list():
                value = [recursive_infer(i) for i in value]
            case dict():
                for key in dict_key_preference:
                    if key in value:
                        value = value[key]
                        value = recursive_infer(value)
                        return value

        return value

    value = recursive_infer(value)
    return value


def generate_directed_linked_issues(issue_fields: dict):
    for issue_link in issue_fields["issuelinks"]:
        for direction in ["inward", "outward"]:
            if linked_issue := issue_link.get(f"{direction}Issue"):
                directed_link = issue_link["type"][direction]
                yield directed_link, linked_issue


def normalize_line_breaks(string: str) -> str:
    string = re.sub(r"[^\S\n]+", " ", string)  # Replace non-newline whitespaces with single space.
    string = re.sub(r"\s*\n\s*", "\n", string)  # Replace multiple newlines with single newline.
    string = string.strip()
    return string


def collect_jira_issue_keys_in_text(text: str) -> list[str]:
    issue_keys = re.findall(r"[a-zA-Z]+-\d+", text)
    issue_keys = sorted({i.upper() for i in issue_keys})
    return issue_keys



from typing import override

from bir_mcp.git_lab.prompts import get_prompts
from bir_mcp.git_lab.sonarqube import SonarQube
from bir_mcp.git_lab.tools import GitLabTools


class GitLab(GitLabTools):
    @override
    def get_prompts(self):
        return get_prompts()



import base64
import inspect
import itertools
import logging
import re
import urllib.parse
from typing import Annotated, override

import fastmcp.tools
import gitlab.v4.objects
import httpx
import pydantic

from bir_mcp.atlassian.utils import collect_jira_issue_keys_in_text
from bir_mcp.core import AhttpxMixin, BaseMcp, build_readonly_tools, build_write_tools
from bir_mcp.git_lab.annotations import BranchType, MergeRequestIidType, ProjectPathOrUrlType
from bir_mcp.git_lab.utils import (
    GitLabUrl,
    MergeRequest,
    aget_project_id_from_path,
    get_ansi_escape_pattern,
    get_gitlab_pat_scopes,
)
from bir_mcp.local.filesystem import find_secrets_in_text
from bir_mcp.utils import (
    araise_for_status,
    filter_dict_by_keys,
    get_line_separated_text_overview,
    prepend_url_path_prefix_if_not_present,
    to_maybe_ssl_context,
)


class GitLabTools(BaseMcp, AhttpxMixin):
    def __init__(
        self,
        private_token: str,
        url: str = "https://gitlab.kapitalbank.az",
        ssl_verify: bool | str = True,
        timezone: str = "UTC",
    ):
        """
        GitLab GraphQL docs: https://docs.gitlab.com/api/graphql/
        Local instance GraphQL explorer: https://gitlab.kapitalbank.az/-/graphql-explorer
        """
        super().__init__(timezone=timezone)
        self.api_version = 4
        self.gitlab = gitlab.Gitlab(
            url=url,
            private_token=private_token,
            ssl_verify=ssl_verify,
            api_version=str(self.api_version),
        )
        self.gitlab.auth()
        self.ahttpx = httpx.AsyncClient(
            base_url=url,
            headers={"PRIVATE-TOKEN": private_token},
            verify=to_maybe_ssl_context(ssl_verify),
            event_hooks={"response": [araise_for_status]},
        )
        self.agraphql = gitlab.AsyncGraphQL(
            url=url,
            token=private_token,
            ssl_verify=ssl_verify,
        )
        self.gitlab_url = GitLabUrl(base_url=url)

    @override
    def get_tag(self):
        return "gitlab"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="Bir GitLab MCP server",
            instructions=inspect.cleandoc("""
                GitLab related tools.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        read_tools = [
            self.get_project_info,
            self.list_all_repo_branch_files,
            self.get_file_content,
            self.search_in_repository,
            self.get_latest_pipeline_info,
            self.get_merge_request_overview,
            self.get_merge_request_overview_from_url,
            self.get_merge_request_file_diffs,
            self.search_users,
            self.search_merge_requests,
        ]
        write_tools_by_scopes = {
            ("api",): [
                self.create_merge_request,
                self.create_branch,
                self.http_request,
            ],
        }
        tools = build_readonly_tools(read_tools, max_output_length=max_output_length, tags=tags)
        pat_scopes = set(get_gitlab_pat_scopes(self.gitlab))
        for scopes, write_tools in write_tools_by_scopes.items():
            if set(scopes) - pat_scopes:
                logging.info(
                    f"GitLab token doesn't have {scopes} scope, skipping tools {write_tools}."
                )
            else:
                tools.extend(
                    build_write_tools(write_tools, max_output_length=max_output_length, tags=tags)
                )

        return tools

    @override
    def prepend_api(self, path: str) -> str:
        return prepend_url_path_prefix_if_not_present(f"api/v{self.api_version}", path)

    def get_project_from_url(self, project_path_or_url: str) -> gitlab.v4.objects.Project:
        project_path = self.gitlab_url.extract_project_path(project_path_or_url)
        project = self.gitlab.projects.get(project_path)
        return project

    async def get_project_id(self, project_path_or_url: ProjectPathOrUrlType) -> int:
        project_path = self.gitlab_url.extract_project_path(project_path_or_url)
        project_id = await aget_project_id_from_path(graphql=self.agraphql, project=project_path)
        return project_id

    async def _paginated_get(self, endpoint: str, per_page: int = 100, **params) -> list[dict]:
        """Helper method to handle paginated GitLab API requests."""
        pages = []
        for page in itertools.count(1):
            response = await self.ahttpx_get(endpoint, per_page=per_page, page=page, **params)
            page = response.json()
            if not page:
                break

            pages.extend(page)

        return pages

    async def get_project_info(self, project_path_or_url: ProjectPathOrUrlType) -> dict:
        """
        Retrieves metadata and general info about a GitLab project identified by the repo url,
        such as name, description, branches, last activity, topics (tags), etc.
        """
        project_id = await self.get_project_id(project_path_or_url)
        project_response = await self.ahttpx_get(f"projects/{project_id}")
        project = project_response.json()

        branches_response = await self.ahttpx_get(f"projects/{project_id}/repository/branches")
        branches = branches_response.json()
        branches = [branch["name"] for branch in branches]

        project_info = filter_dict_by_keys(
            project, ["path_with_namespace", "topics", "description", "web_url"]
        )
        project_info |= {
            "last_activity_at": self.format_datetime_for_ai(project["last_activity_at"]),
            "branches": branches,
        }
        return project_info

    async def list_all_repo_branch_files(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        branch: BranchType,
    ) -> dict:
        """Recursively lists all files and directories in the repository."""
        project_id = await self.get_project_id(project_path_or_url)
        tree = await self._paginated_get(
            f"projects/{project_id}/repository/tree", ref=branch, recursive="true"
        )
        tree = {"files": [filter_dict_by_keys(item, ["path", "type"]) for item in tree]}
        return tree

    async def get_file_content(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        branch: BranchType,
        file_path: Annotated[
            str,
            pydantic.Field(
                description="The path to the file relative to the root of the repository."
            ),
        ],
    ) -> dict:
        """Retrieves the text content of a specific file."""
        project_id = await self.get_project_id(project_path_or_url)
        encoded_file_path = urllib.parse.quote(file_path, safe="")
        response = await self.ahttpx_get(
            f"projects/{project_id}/repository/files/{encoded_file_path}", ref=branch
        )
        file = response.json()
        content = base64.b64decode(file["content"]).decode()
        return {"file_content": content}

    async def search_in_repository(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        branch: BranchType,
        query: Annotated[str, pydantic.Field(description="The text query to search for.")],
    ) -> dict:
        """
        Performs a basic search for a text query within the repo's files.
        Doesn't support regex, but is case-insensitive.
        Returns a list of occurences within files, with file path, starting line in the file
        and a snippet of the contextual window in which the query was found.
        For details see the [API docs](https://docs.gitlab.com/api/search/#project-search-api).
        """
        project_id = await self.get_project_id(project_path_or_url)
        response = await self.ahttpx_get(
            f"projects/{project_id}/search", scope="blobs", search=query, ref=branch
        )
        search_results = response.json()
        search_results = [
            {
                "file_path": result["path"],
                "starting_line_in_file": result["startline"],
                "snippet": result["data"],
            }
            for result in search_results
        ]
        return {"search_results": search_results}

    async def get_latest_pipeline_info(self, project_path_or_url: ProjectPathOrUrlType) -> dict:
        """Retrieves the latest pipeline info, such as url, status, duration, commit, jobs, etc."""
        project_id = await self.get_project_id(project_path_or_url)
        pipelines = await self.ahttpx_get(
            f"projects/{project_id}/pipelines", per_page=1, sort="desc"
        )
        pipelines = pipelines.json()
        if not pipelines:
            return {"error": "No pipelines found"}

        pipeline = pipelines[0]
        pipeline_id = pipeline["id"]
        pipeline_commit = pipeline["sha"]

        commit = await self.ahttpx_get(
            f"projects/{project_id}/repository/commits/{pipeline_commit}"
        )
        commit = filter_dict_by_keys(
            commit.json(),
            ["id", "title", "author_name", "web_url"],
        )

        jobs = await self.ahttpx_get(f"projects/{project_id}/pipelines/{pipeline_id}/jobs")
        jobs = [
            filter_dict_by_keys(
                job,
                ["name", "status", "stage", "allow_failure", "web_url"],
            )
            for job in jobs.json()
        ]
        pipeline_info = filter_dict_by_keys(pipeline, ["web_url", "status", "source"])
        pipeline_info |= {
            "created_at": self.format_datetime_for_ai(pipeline["created_at"]),
            "commit": commit,
            "jobs": jobs,
        }
        return pipeline_info

    async def get_merge_request_file_diffs(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        merge_request_iid: MergeRequestIidType,
        file_paths: Annotated[
            list[str],
            pydantic.Field(description="The paths to the new files relative to the merge request."),
        ],
    ) -> dict:
        """
        Fetch full diffs for specific files in a merge request, based on the new file paths.
        Should be called with several paths at once for efficiency.
        """
        merge_request = await self.get_merge_request(project_path_or_url, merge_request_iid)
        diffs = []
        for change in merge_request["changes"]:
            if change["new_path"] in file_paths:
                diff = filter_dict_by_keys(change, ["new_path", "diff"])
                diffs.append(diff)

        diffs = {"diffs": diffs}
        return diffs

    async def get_merge_request(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        merge_request_iid: MergeRequestIidType,
    ) -> dict:
        project_id = await self.get_project_id(project_path_or_url)
        endpoint = f"/projects/{project_id}/merge_requests/{merge_request_iid}/changes"
        response = await self.ahttpx_get(endpoint, access_raw_diffs=True, unidiff=True)
        merge_request = response.json()
        return merge_request

    async def get_merge_request_overview(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        merge_request_iid: MergeRequestIidType,
        first_diff_lines: int = 6,
    ) -> dict:
        """
        Fetch overview details about a GitLab merge request, such as title, description, branches, diffs,
        pipeline status and jobs. The diffs and job traces are truncated, for full diffs and files use
        other tools.
        """
        merge_request = await self.get_merge_request(project_path_or_url, merge_request_iid)
        overview = filter_dict_by_keys(
            merge_request, ["title", "description", "source_branch", "target_branch"]
        )
        head_pipeline = merge_request["head_pipeline"]
        overview["pipeline_status"] = (head_pipeline or {}).get("status")
        jobs = []
        if head_pipeline:
            project_id = merge_request["project_id"]
            endpoint = f"/projects/{project_id}/pipelines/{head_pipeline['id']}/jobs"
            response = await self.ahttpx_get(endpoint, per_page=100)
            ansi = re.compile(get_ansi_escape_pattern())
            for job in response.json():
                job_id = job["id"]
                trace = await self.ahttpx_get(f"/projects/{project_id}/jobs/{job_id}/trace")
                trace = trace.text
                trace = ansi.sub(repl="", string=trace)
                if job["name"] == "Run Up: PreFlight":
                    if match := re.search("\nJIRA_NAMES: *(.*?)\n", trace):
                        jira_issues = match.group(1)
                        jira_issues = collect_jira_issue_keys_in_text(jira_issues)
                        overview["related_jira_issues"] = jira_issues

                trace_overview = get_line_separated_text_overview(
                    trace, line_chunks_len=first_diff_lines
                )
                job = filter_dict_by_keys(job, ["name", "stage", "status", "web_url"])
                job["trace_overview"] = trace_overview
                jobs.append(job)

        overview["pipeline_jobs"] = jobs
        changes_overview = []
        for change in merge_request["changes"]:
            change_overview = filter_dict_by_keys(change, ["old_path", "new_path"])
            change_overview["diff_first_lines"] = "\n".join(
                change["diff"].splitlines()[:first_diff_lines]
            )
            if change["new_file"]:
                change_overview.pop("old_path")
            if change["deleted_file"]:
                change_overview.pop("new_path")

            detected_secrets = find_secrets_in_text(change["diff"])
            if detected_secrets["detected_secrets"]:
                change_overview |= detected_secrets

            changes_overview.append(change_overview)

        overview["changes_overview"] = changes_overview
        return overview

    async def get_merge_request_overview_from_url(
        self,
        merge_request_url: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The url for merge request, in the format:
                    "https://{gitlab_domain}/{project_path}/-/merge_requests/{merge_request_iid}".
                """)
            ),
        ],
    ) -> dict:
        """
        Fetch overview about a GitLab merge request, is a convenience function that calls the
        get_merge_request_overview tool.
        """
        merge_request: MergeRequest = self.gitlab_url.extract_merge_request(merge_request_url)
        merge_request_overview = await self.get_merge_request_overview(
            project_path_or_url=merge_request.project_path,
            merge_request_iid=merge_request.merge_request_iid,
        )
        return merge_request_overview

    async def create_merge_request(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        source_branch: Annotated[
            str,
            pydantic.Field(
                description="The source branch name for the merge request, usually a feature or testing branch."
            ),
        ],
        target_branch: Annotated[
            str,
            pydantic.Field(
                description="The target branch name for the merge request, usually 'testing' or 'main'."
            ),
        ],
        title: Annotated[
            str,
            pydantic.Field(description="The title of the merge request."),
        ],
        description: Annotated[
            str,
            pydantic.Field(description="The description of the merge request."),
        ],
        assignee_id: Annotated[
            int | None,
            pydantic.Field(description="The GitLab user id to assign the merge request to."),
        ] = None,
        labels: Annotated[
            list[str] | None,
            pydantic.Field(description="List of labels to add to the merge request."),
        ] = None,
        remove_source_branch: Annotated[
            bool,
            pydantic.Field(
                description="Whether to remove the source branch when the merge request is merged."
            ),
        ] = True,
        squash: Annotated[
            bool,
            pydantic.Field(description="Whether to squash commits when merging."),
        ] = False,
    ) -> dict:
        """Create a new GitLab merge request."""
        project_id = await self.get_project_id(project_path_or_url)
        merge_request = {
            "source_branch": source_branch,
            "target_branch": target_branch,
            "title": title,
            "remove_source_branch": remove_source_branch,
            "squash": squash,
            "description": description,
        }
        if assignee_id:
            merge_request["assignee_id"] = assignee_id
        if labels:
            merge_request["labels"] = ",".join(labels)

        merge_request = await self.ahttpx_post(
            f"projects/{project_id}/merge_requests", json=merge_request
        )
        merge_request = filter_dict_by_keys(merge_request.json(), ["id", "iid", "web_url"])
        return {"created_merge_request": merge_request}

    async def search_users(
        self,
        query: Annotated[
            str,
            pydantic.Field(description="The query to search users by the username, name or email."),
        ],
    ) -> dict:
        """Search for GitLab users, primarily used to get user emails or internal ids."""
        # Only admin can search by private emails and get other user's private emails.
        users = await self.ahttpx_get("users", search=query)
        users = [
            filter_dict_by_keys(user, ["id", "username", "name", "email"]) for user in users.json()
        ]
        return {"users": users}

    async def create_branch(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        new_branch_name: str,
        source_ref: Annotated[
            str,
            pydantic.Field(description="The source branch, tag, or commit SHA."),
        ],
    ) -> dict:
        """Creates a new branch in the GitLab project."""
        project_id = await self.get_project_id(project_path_or_url)
        branch = {"branch": new_branch_name, "ref": source_ref}
        branch = await self.ahttpx_post(f"projects/{project_id}/repository/branches", json=branch)
        branch = branch.json()
        return {"new_branch_url": branch["web_url"]}

    async def create_file(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        branch: str,
        file_path: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The path to the file relative to the root of the repository. 
                    Any parent directories will be created if they don't exist.
                """)
            ),
        ],
        file_content: str,
        author_email: str,
        commit_message: str | None = None,
    ) -> dict:
        """Creates a new file in the GitLab project branch. Fails if file already exists."""
        raise_if_branch_editing_not_allowed(branch)
        project_id = await self.get_project_id(project_path_or_url)

        commit_message = commit_message or f"Add file {file_path} to {branch} branch"
        actions = [
            {
                "action": "create",
                "file_path": file_path,
                "content": file_content,
                "encoding": "text",
            }
        ]
        commit = {
            "branch": branch,
            "commit_message": commit_message,
            "author_email": author_email,
            "actions": actions,
        }
        commit = await self.ahttpx_post(f"projects/{project_id}/repository/commits", json=commit)
        commit = commit.json()
        return {"commit_url": commit["web_url"]}

    async def search_merge_requests(
        self,
        project_path_or_url: ProjectPathOrUrlType,
        source_branch: str = "testing",
        target_branch: str = "master",
        state: str = "opened",
    ) -> dict:
        """Search for merge requests by source branch, target branch and state."""
        project_id = await self.get_project_id(project_path_or_url)
        merge_requests = await self.ahttpx_get(
            f"projects/{project_id}/merge_requests",
            source_branch=source_branch,
            target_branch=target_branch,
            state=state,
        )
        merge_requests = [
            filter_dict_by_keys(mr, ["iid", "title", "description", "web_url"])
            for mr in merge_requests.json()
        ]
        return {"merge_requests": merge_requests}


def raise_if_branch_editing_not_allowed(branch: str):
    if not branch.startswith("feature/"):
        raise ValueError(
            "Due to safety considerations, modifying branches or content is only allowed in feature branches."
        )



import re
import string

import gitlab as gl
import httpx
import pydantic

from bir_mcp.utils import filter_dict_by_keys, value_to_key


class GraphqlGlobalId(pydantic.BaseModel):
    """
    GitLab GraphQL often uses global ids for object identifiers,
    see https://docs.gitlab.com/api/graphql/#global-ids
    """

    global_id: str
    schema: str
    provider: str
    path: list[str]
    entity_id: int

    @staticmethod
    def from_global_id(global_id: str):
        schema, path = global_id.split("://")
        provider, *path, entity_id = path.split("/")
        global_id = GraphqlGlobalId(
            global_id=global_id,
            schema=schema,
            provider=provider,
            path=path,
            entity_id=int(entity_id),
        )
        return global_id


async def aget_project_id_from_path(graphql: gl.AsyncGraphQL, project: str) -> int:
    query = """
        query {
            project(fullPath: "$project") {
                id
            }
        }
    """
    result = await aexecute_parametrized_graphql_query(graphql, query, project=project)
    global_id = result["project"]["id"]
    global_id = GraphqlGlobalId.from_global_id(global_id)
    return global_id.entity_id


class MergeRequest(pydantic.BaseModel):
    project_path: str
    merge_request_iid: int


class GitLabUrl:
    def __init__(self, base_url: str):
        self.base_url = base_url.strip("/")
        re_base_url = re.escape(self.base_url)
        project_path = rf"({re_base_url}/)?(?P<project_path>[\w\-/]+)"
        merge_request = rf"{project_path}/-/merge_requests/(?P<merge_request_iid>\d+)"
        self.project_path_pattern = re.compile(project_path)
        self.merge_request_pattern = re.compile(merge_request)

    def extract_project_path(self, project_path_or_url: str) -> str:
        match = self.project_path_pattern.match(project_path_or_url)
        if not match:
            raise ValueError(
                f"String '{project_path_or_url}' does not fit expected pattern for GitLab project:\n"
                f"{self.project_path_pattern.pattern}"
            )

        project_path = match.groupdict()["project_path"].strip("/")
        return project_path

    def extract_merge_request(self, merge_request_url: str) -> MergeRequest:
        match = self.merge_request_pattern.match(merge_request_url)
        if not match:
            raise ValueError(
                f"String '{merge_request_url}' does not fit expected pattern for GitLab merge request:\n"
                f"{self.merge_request_pattern.pattern}"
            )

        merge_request = MergeRequest(**match.groupdict())
        return merge_request


def get_merge_request_diffs(gitlab: gl.Gitlab, project: str, merge_request_iid: int) -> dict:
    project = gitlab.projects.get(project)
    merge_request = project.mergerequests.get(merge_request_iid)
    # https://docs.gitlab.com/api/merge_requests/#get-single-merge-request-changes
    changes = merge_request.changes(access_raw_diffs=True, unidiff=True)
    changes = [
        filter_dict_by_keys(change, ["old_path", "new_path", "diff"])
        for change in changes["changes"]
    ]
    merge_request = filter_dict_by_keys(merge_request.asdict(), ["title", "description"])
    merge_request |= {"changes": changes}
    return merge_request


async def aget_merge_request_data(
    project: str,
    merge_request_iid: int,
    agraphql: gl.AsyncGraphQL,
    ahttpx_client: httpx.AsyncClient,
    access_raw_diffs: bool = True,
    unidiff: bool = True,
    api_version: int = 4,
) -> dict:
    """https://docs.gitlab.com/api/merge_requests/#get-single-merge-request-changes"""
    project_id = await aget_project_id_from_path(agraphql, project)
    changes_url = (
        f"api/v{api_version}/projects/{project_id}/merge_requests/{merge_request_iid}/changes"
    )
    response = await ahttpx_client.get(
        changes_url, params=dict(access_raw_diffs=access_raw_diffs, unidiff=unidiff)
    )
    merge_request = response.json()
    merge_request = filter_dict_by_keys(merge_request, ["title", "description", "changes"])
    merge_request["changes"] = [
        filter_dict_by_keys(change, ["old_path", "new_path", "diff"])
        for change in merge_request["changes"]
    ]
    return merge_request


async def aexecute_parametrized_graphql_query(
    graphql: gl.AsyncGraphQL, query: str, **parameters
) -> dict:
    query = string.Template(query).substitute(parameters)
    result = await graphql.execute(query)
    return result


async def aget_all_schema_types(graphql: gl.AsyncGraphQL) -> dict[str, str]:
    query = """
        query {
            __schema {
                types {
                    name
                    kind
                }
            }
        }
    """
    result = await graphql.execute(query)
    types = value_to_key(result["__schema"]["types"], "name")
    return types


async def aget_all_entity_type_fields(
    graphql: gl.AsyncGraphQL, entity_type: str, include_datatype_metadata: bool = False
) -> dict:
    """Note that entity_type is case-sensitive."""
    datatype_metadata = """
        type {
            name
            kind
            ofType {
                name
                kind
            }
        }
    """
    query = """
        query {
        __type(name: "$entity_type") {
            fields {
                name
                description
                $datatype_metadata
                args {
                    name
                    description
                    $datatype_metadata
                }
            }
        }
    }
    """
    result = await aexecute_parametrized_graphql_query(
        graphql,
        query,
        entity_type=entity_type,
        datatype_metadata=datatype_metadata if include_datatype_metadata else "",
    )
    fields = result["__type"]["fields"]
    for field in fields:
        field["args"] = value_to_key(field["args"], "name")

    fields = value_to_key(fields, "name")
    return fields


async def aget_merge_request_commits(
    graphql: gl.AsyncGraphQL, project: str, merge_request_iid: int, first: int = 1
) -> list:
    query = """
        query {
            project(fullPath: "$project") {
                mergeRequest(iid: "$merge_request_iid") {
                    commits(first: $first) {
                        nodes {
                            title
                            message
                            diffs {
                                oldPath
                                newPath
                                diff
                            }
                        }
                    }
                }
            }
        }
    """
    result = await aexecute_parametrized_graphql_query(
        graphql, query, project=project, merge_request_iid=merge_request_iid, first=first
    )
    commits = result["project"]["mergeRequest"]["commits"]["nodes"]
    return commits


def get_ansi_escape_pattern() -> str:
    return r"\x1b\[[;\d]*[A-Za-z]"


def get_gitlab_pat_scopes(gitlab: gl.Gitlab) -> list[str]:
    pat = gitlab.personal_access_tokens.get("self")
    scopes = sorted(pat.scopes)
    return scopes



"""Tests for GitLab MCP tools."""

import asyncio
import base64
from unittest.mock import Mock, patch

from bir_mcp.git_lab.prompts import get_feature_development_prompt, get_merge_request_review_prompt


class TestGitLabToolsReadOnly:
    """Test read-only GitLab tools."""

    def test_get_project_info(self, gitlab_tools):
        """Test get_project_info tool."""
        expected_result = {
            "name_with_namespace": "test/project",
            "topics": ["python", "testing"],
            "description": "Test project",
            "web_url": "https://gitlab.example.com/test/project",
            "branches": ["main"],
            "last_activity_at": "2024-01-01 00:00:00 UTC",
        }

        # Mock the async method
        async def mock_async_method(*args, **kwargs):
            return expected_result

        gitlab_tools.get_project_info = mock_async_method

        # Since we're testing the async method, we need to run it
        result = asyncio.run(gitlab_tools.get_project_info("test/project"))

        assert result["name_with_namespace"] == "test/project"
        assert result["topics"] == ["python", "testing"]
        assert result["description"] == "Test project"
        assert result["web_url"] == "https://gitlab.example.com/test/project"
        assert result["branches"] == ["main"]
        assert "last_activity_at" in result

    def test_list_all_repo_branch_files(self, gitlab_tools):
        """Test list_all_repo_branch_files tool."""
        expected_result = {
            "files": [
                {"path": "README.md", "type": "blob"},
                {"path": "src/main.py", "type": "blob"},
                {"path": "src", "type": "tree"},
            ]
        }

        # Mock the async method
        async def mock_async_method(*args, **kwargs):
            return expected_result

        gitlab_tools.list_all_repo_branch_files = mock_async_method

        # Since we're testing the async method, we need to run it
        result = asyncio.run(gitlab_tools.list_all_repo_branch_files("test/project", "main"))

        assert "files" in result
        assert len(result["files"]) == 3
        assert result["files"][0] == {"path": "README.md", "type": "blob"}

    def test_get_file_content(self, gitlab_tools, mock_project):
        """Test get_file_content tool."""
        file_content = "print('Hello World')"
        encoded_content = base64.b64encode(file_content.encode()).decode()

        mock_file_data = {"content": encoded_content}

        # Mock async methods
        async def mock_get_project_id(project_path_or_url):
            return 123

        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_file_data
            return mock_response

        with (
            patch.object(gitlab_tools, "get_project_id", side_effect=mock_get_project_id),
            patch.object(gitlab_tools, "ahttpx_get", side_effect=mock_ahttpx_get),
        ):
            result = asyncio.run(
                gitlab_tools.get_file_content("test/project", "main", "src/main.py")
            )

            assert result["file_content"] == file_content

    def test_search_in_repository(self, gitlab_tools, mock_project):
        """Test search_in_repository tool."""
        mock_search_results = [
            {
                "path": "src/main.py",
                "data": "def main():\n    print('Hello')",
                "startline": 1,
            }
        ]
        expected_results = [
            {
                "file_path": "src/main.py",
                "starting_line_in_file": 1,
                "snippet": "def main():\n    print('Hello')",
            }
        ]

        # Mock async methods
        async def mock_get_project_id(project_path_or_url):
            return 123

        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_search_results
            return mock_response

        with (
            patch.object(gitlab_tools, "get_project_id", side_effect=mock_get_project_id),
            patch.object(gitlab_tools, "ahttpx_get", side_effect=mock_ahttpx_get),
        ):
            result = asyncio.run(gitlab_tools.search_in_repository("test/project", "main", "Hello"))

            assert result["search_results"] == expected_results

    def test_get_latest_pipeline_info(self, gitlab_tools, mock_project):
        """Test get_latest_pipeline_info tool."""
        mock_pipeline_data = {
            "id": 123,
            "web_url": "https://gitlab.example.com/test/project/-/pipelines/123",
            "created_at": "2024-01-01T00:00:00Z",
            "status": "success",
            "source": "push",
            "duration": 300,
            "queued_duration": 30,
            "sha": "abc123",
        }

        mock_commit_data = {
            "id": "abc123",
            "title": "Test commit",
            "author_name": "Test Author",
            "web_url": "https://gitlab.example.com/test/project/-/commit/abc123",
        }

        mock_jobs_data = [
            {
                "name": "test",
                "status": "success",
                "stage": "test",
                "allow_failure": False,
                "web_url": "https://gitlab.example.com/test/project/-/jobs/456",
            }
        ]

        # Mock async methods
        async def mock_get_project_id(project_path_or_url):
            return 123

        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            if "pipelines" in endpoint and "jobs" not in endpoint:
                mock_response.json.return_value = [mock_pipeline_data]
            elif "commits" in endpoint:
                mock_response.json.return_value = mock_commit_data
            elif "jobs" in endpoint:
                mock_response.json.return_value = mock_jobs_data
            return mock_response

        with (
            patch.object(gitlab_tools, "get_project_id", side_effect=mock_get_project_id),
            patch.object(gitlab_tools, "ahttpx_get", side_effect=mock_ahttpx_get),
        ):
            result = asyncio.run(gitlab_tools.get_latest_pipeline_info("test/project"))

            assert result["web_url"] == "https://gitlab.example.com/test/project/-/pipelines/123"
            assert result["status"] == "success"
            assert result["commit"]["title"] == "Test commit"
            assert len(result["jobs"]) == 1
            assert result["jobs"][0]["name"] == "test"

    @patch("asyncio.run")
    def test_get_merge_request_overview(
        self, mock_asyncio_run, gitlab_tools, mock_project, mock_merge_request
    ):
        """Test get_merge_request_overview tool."""
        expected_result = {
            "merge_request": {"id": 456, "title": "Test MR"},
            "pipeline_status": "success",
            "file_changes": [{"old_path": "file1.py", "new_path": "file1.py"}],
        }
        mock_asyncio_run.return_value = expected_result

        # Mock the async method
        async def mock_async_method(*args, **kwargs):
            return expected_result

        gitlab_tools.get_merge_request_overview = mock_async_method

        # Since we're testing the async method, we need to run it
        result = asyncio.run(gitlab_tools.get_merge_request_overview("test/project", 1))

        assert result["merge_request"]["id"] == 456
        assert result["merge_request"]["title"] == "Test MR"
        assert result["pipeline_status"] == "success"
        assert len(result["file_changes"]) == 1

    def test_get_merge_request_overview_from_url(self, gitlab_tools):
        """Test get_merge_request_overview_from_url tool."""
        mr_url = "https://gitlab.example.com/test/project/-/merge_requests/1"
        expected_result = {"merge_request": {"id": 456}}

        # Mock the async method
        async def mock_async_method(*args, **kwargs):
            return expected_result

        gitlab_tools.get_merge_request_overview_from_url = mock_async_method

        # Since we're testing the async method, we need to run it
        result = asyncio.run(gitlab_tools.get_merge_request_overview_from_url(mr_url))

        assert result["merge_request"]["id"] == 456

    def test_get_merge_request_file_diffs(self, gitlab_tools, mock_project, mock_merge_request):
        """Test get_merge_request_file_diffs tool."""
        expected_result = {
            "file_diffs": [
                {"new_path": "file1.py", "diff": "@@ -1,3 +1,3 @@\n-old line\n+new line"},
                {
                    "new_path": "file2.py",
                    "diff": "@@ -1,2 +1,2 @@\n-another old line\n+another new line",
                },
            ]
        }

        # Mock the async method
        async def mock_async_method(*args, **kwargs):
            return expected_result

        gitlab_tools.get_merge_request_file_diffs = mock_async_method

        # Since we're testing the async method, we need to run it
        result = asyncio.run(
            gitlab_tools.get_merge_request_file_diffs("test/project", 1, ["file1.py", "file2.py"])
        )

        assert len(result["file_diffs"]) == 2
        assert result["file_diffs"][0]["new_path"] == "file1.py"
        assert "diff" in result["file_diffs"][0]

    def test_search_users(self, gitlab_tools, mock_user):
        """Test search_users tool."""
        mock_users_data = [{
            "id": 123,
            "username": "testuser",
            "name": "Test User",
            "email": "test@example.com"
        }]
        
        # Mock async method
        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_users_data
            return mock_response

        with patch.object(gitlab_tools, "ahttpx_get", side_effect=mock_ahttpx_get):
            result = asyncio.run(gitlab_tools.search_users("testuser"))

            assert len(result["users"]) == 1
            assert result["users"][0]["username"] == "testuser"
            assert result["users"][0]["email"] == "test@example.com"

    def test_search_merge_requests(self, gitlab_tools, mock_project, mock_merge_request):
        """Test search_merge_requests tool."""
        expected_result = {"merge_requests": [{"title": "Test MR", "id": 456}]}

        # Mock the async method
        async def mock_async_method(*args, **kwargs):
            return expected_result

        gitlab_tools.search_merge_requests = mock_async_method

        # Since we're testing the async method, we need to run it
        result = asyncio.run(
            gitlab_tools.search_merge_requests(
                "test/project", source_branch="feature", target_branch="main", state="opened"
            )
        )

        assert len(result["merge_requests"]) == 1
        assert result["merge_requests"][0]["title"] == "Test MR"


class TestGitLabToolsWrite:
    """Test write GitLab tools."""

    def test_create_merge_request(self, gitlab_tools, mock_project):
        """Test create_merge_request tool."""
        mock_mr_data = {
            "id": 789,
            "iid": 2,
            "web_url": "https://gitlab.example.com/test/project/-/merge_requests/2"
        }
        
        # Mock async methods
        async def mock_get_project_id(project_path_or_url):
            return 123
            
        async def mock_ahttpx_post(url, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_mr_data
            return mock_response

        with patch.object(gitlab_tools, "get_project_id", side_effect=mock_get_project_id), \
             patch.object(gitlab_tools, "ahttpx") as mock_ahttpx:
            mock_ahttpx.post = mock_ahttpx_post
            
            result = asyncio.run(gitlab_tools.create_merge_request(
                project_path_or_url="test/project",
                source_branch="feature-branch",
                target_branch="main",
                title="New Feature",
                description="Adding new feature",
                assignee_id=123,
                labels=["enhancement", "feature"],
            ))

            assert result["created_merge_request"]["id"] == 789
            assert result["created_merge_request"]["iid"] == 2

    def test_create_branch(self, gitlab_tools, mock_project):
        """Test create_branch tool."""
        mock_branch_data = {
            "web_url": "https://gitlab.example.com/test/project/-/tree/new-feature"
        }
        
        # Mock async methods
        async def mock_get_project_id(project_path_or_url):
            return 123
            
        async def mock_ahttpx_post(url, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_branch_data
            return mock_response

        with patch.object(gitlab_tools, "get_project_id", side_effect=mock_get_project_id), \
             patch.object(gitlab_tools, "ahttpx") as mock_ahttpx:
            mock_ahttpx.post = mock_ahttpx_post
            
            result = asyncio.run(gitlab_tools.create_branch("test/project", "new-feature", "main"))

            assert (
                result["new_branch_url"]
                == "https://gitlab.example.com/test/project/-/tree/new-feature"
            )

    @patch("bir_mcp.git_lab.tools.httpx.AsyncClient")
    def test_http_request(self, mock_httpx, gitlab_tools):
        """Test http_request tool."""
        mock_response = Mock()
        mock_response.json.return_value = {"result": "success"}

        # Mock the async client context manager
        mock_client = Mock()
        mock_client.request.return_value = mock_response
        mock_httpx.return_value.__aenter__.return_value = mock_client

        # Since this is an async method, we need to test it differently
        # For now, we'll just test that the method exists and has the right signature
        assert hasattr(gitlab_tools, "http_request")

        # Test the method signature
        import inspect

        sig = inspect.signature(gitlab_tools.http_request)
        assert "path" in sig.parameters
        assert "http_method" in sig.parameters
        assert "json_query_params" in sig.parameters
        assert "content" in sig.parameters


class TestGitLabToolsPrompts:
    """Test GitLab prompt functions."""

    def test_get_feature_development_prompt(self):
        """Test get_feature_development_prompt function."""
        result = get_feature_development_prompt("PROJ", "Add user authentication")

        assert isinstance(result, str)
        assert "Add user authentication" in result
        assert "PROJ" in result
        assert "feature_description" in result

    def test_get_merge_request_review_prompt(self):
        """Test get_merge_request_review_prompt function."""
        mr_url = "https://gitlab.example.com/test/project/-/merge_requests/1"
        result = get_merge_request_review_prompt(mr_url)

        assert isinstance(result, str)
        assert mr_url in result
        assert "review" in result.lower()


class TestGitLabToolsHelpers:
    """Test GitLab helper methods."""

    def test_get_project_from_url(self, gitlab_tools):
        """Test get_project_from_url helper method."""
        mock_project = Mock()
        gitlab_tools.gitlab.projects.get.return_value = mock_project

        # The actual method calls gitlab.projects.get with the path directly
        result = gitlab_tools.get_project_from_url("test/project")

        assert result == mock_project
        gitlab_tools.gitlab.projects.get.assert_called_once_with("test/project")

    def test_get_mcp_tools_returns_all_tools(self, gitlab_tools):
        """Test that get_mcp_tools returns all expected tools."""
        with patch("bir_mcp.git_lab.tools.get_gitlab_pat_scopes") as mock_scopes:
            mock_scopes.return_value = ["api"]

            with (
                patch("bir_mcp.git_lab.tools.build_readonly_tools") as mock_readonly,
                patch("bir_mcp.git_lab.tools.build_write_tools") as mock_write,
            ):
                mock_readonly.return_value = ["read_tool1", "read_tool2"]
                mock_write.return_value = ["write_tool1", "write_tool2"]

                gitlab_tools.get_mcp_tools()

                # Verify read tools were called with correct functions
                read_tools_arg = mock_readonly.call_args[0][0]
                assert gitlab_tools.get_project_info in read_tools_arg
                assert gitlab_tools.list_all_repo_branch_files in read_tools_arg
                assert gitlab_tools.get_file_content in read_tools_arg
                assert gitlab_tools.search_in_repository in read_tools_arg
                assert gitlab_tools.get_latest_pipeline_info in read_tools_arg
                assert gitlab_tools.get_merge_request_overview in read_tools_arg
                assert gitlab_tools.get_merge_request_overview_from_url in read_tools_arg
                assert gitlab_tools.get_merge_request_file_diffs in read_tools_arg
                assert gitlab_tools.search_users in read_tools_arg
                assert gitlab_tools.search_merge_requests in read_tools_arg

                # Verify write tools were called with correct functions
                write_tools_arg = mock_write.call_args[0][0]
                assert gitlab_tools.create_merge_request in write_tools_arg
                assert gitlab_tools.create_branch in write_tools_arg
                assert gitlab_tools.http_request in write_tools_arg

    def test_get_mcp_tools_skips_write_tools_without_scopes(self, gitlab_tools):
        """Test that get_mcp_tools skips write tools when PAT doesn't have required scopes."""
        with patch("bir_mcp.git_lab.tools.get_gitlab_pat_scopes") as mock_scopes:
            mock_scopes.return_value = []  # No scopes

            with (
                patch("bir_mcp.git_lab.tools.build_readonly_tools") as mock_readonly,
                patch("bir_mcp.git_lab.tools.build_write_tools") as mock_write,
            ):
                mock_readonly.return_value = ["read_tool1", "read_tool2"]

                gitlab_tools.get_mcp_tools()

                # Verify only read tools were built
                mock_readonly.assert_called_once()
                mock_write.assert_not_called()



import inspect
from typing import Annotated

import pydantic

ProjectPathOrUrlType = Annotated[
    str,
    pydantic.Field(
        description=inspect.cleandoc("""
            Either the filesystem-like path to a GitLab project with variable depth, for example:
            "organization/project_name" or "organization/namespace/subgroup/project_name",
            or a full url to a GitLab project in the format:
            "https://{gitlab_domain}/{project_path}/[.git]".
        """)
    ),
]
BranchType = Annotated[
    str,
    pydantic.Field(description="The branch name to fetch files from."),
]
MergeRequestIidType = Annotated[
    int,
    pydantic.Field(
        description=inspect.cleandoc("""
                The internal id of a GitLab merge request, can be extracted from a merge request url
                if it ends in the following suffix: "/-/merge_requests/{merge_request_iid}".
            """)
    ),
]



"""
Comprehensive tests for Atlassian MCP tools.
"""

import asyncio
from unittest.mock import Mock, patch

import pytest

from bir_mcp.atlassian import Atlassian


class TestAtlassianTools:
    """Test Atlassian MCP tools."""

    @pytest.fixture
    def atlassian_tools(self):
        """Create Atlassian tools instance for testing."""
        with patch('bir_mcp.atlassian.atlassian.Jira') as mock_jira_class:
            # Mock the Jira class constructor
            mock_jira_instance = Mock()
            mock_jira_instance.url = "https://jira.example.com"
            mock_jira_class.return_value = mock_jira_instance
            
            # Mock search_users
            mock_jira_instance.search_users.return_value = [
                {
                    "accountId": "123",
                    "displayName": "Test User",
                    "emailAddress": "test@example.com"
                }
            ]
            
            # Mock create_issue
            created_issue_mock = Mock()
            created_issue_mock.key = "TEST-456"
            mock_jira_instance.create_issue.return_value = created_issue_mock
            
            with patch('bir_mcp.atlassian.httpx.AsyncClient'):
                tools = Atlassian(
                    token="test-token",
                    url="https://jira.example.com",
                    api_version=3
                )
                return tools

    @pytest.fixture
    def mock_jira_issue_data(self):
        """Mock Jira issue data."""
        return {
            "key": "TEST-123",
            "fields": {
                "summary": "Test Issue",
                "description": "Test description",
                "status": {"name": "Open"},
                "assignee": {"displayName": "Test User"},
                "priority": {"name": "High"},
                "issuetype": {"name": "Bug"},
                "created": "2024-01-01T00:00:00.000+0000",
                "updated": "2024-01-02T00:00:00.000+0000",
                "comment": {"comments": []},
                "subtasks": [],
                "issuelinks": [],
                "parent": None
            }
        }

    @pytest.fixture
    def mock_project_data(self):
        """Mock Jira project data."""
        return {
            "key": "TEST",
            "name": "Test Project",
            "issueTypes": [
                {
                    "name": "Bug",
                    "description": "A problem which impairs or prevents the functions of the product.",
                    "subtask": False,
                },
                {"name": "Task", "description": "A task that needs to be done.", "subtask": False},
                {"name": "Sub-task", "description": "The sub-task of the issue", "subtask": True},
            ],
        }


class TestAtlassianReadOnlyTools(TestAtlassianTools):
    """Test read-only Atlassian tools."""

    def test_get_allowed_issue_types_for_project(self, atlassian_tools, mock_project_data):
        """Test get_allowed_issue_types_for_project tool."""

        # Mock async method
        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_project_data
            return mock_response

        with patch.object(atlassian_tools, "ahttpx_get", side_effect=mock_ahttpx_get):
            result = asyncio.run(atlassian_tools.get_allowed_issue_types_for_project("TEST"))

            assert "tasks" in result
            assert "subtasks" in result
            assert "Bug" in result["tasks"]
            assert "Task" in result["tasks"]
            assert "Sub-task" in result["subtasks"]
            assert (
                result["tasks"]["Bug"]
                == "A problem which impairs or prevents the functions of the product."
            )

    def test_get_raw_jira_issue_fields(self, atlassian_tools, mock_jira_issue_data):
        """Test get_raw_jira_issue_fields tool."""

        # Mock async method
        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_jira_issue_data
            return mock_response

        with patch.object(atlassian_tools, "ahttpx_get", side_effect=mock_ahttpx_get):
            result = asyncio.run(
                atlassian_tools.get_raw_jira_issue_fields("TEST-123", ["summary", "description"])
            )

            assert result["summary"] == "Test Issue"
            assert result["description"] == "Test description"

    def test_get_jira_issue_overview(self, atlassian_tools, mock_jira_issue_data):
        """Test get_jira_issue_overview tool."""

        # Mock async method
        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_jira_issue_data
            return mock_response

        with patch.object(atlassian_tools, "ahttpx_get", side_effect=mock_ahttpx_get):
            with patch('bir_mcp.atlassian.infer_jira_field_representation', side_effect=lambda x: x):
                result = asyncio.run(atlassian_tools.get_jira_issue_overview("TEST-123"))

                assert result["summary"] == "Test Issue"
                assert result["description"] == "Test description"
                assert result["priority"] == {"name": "High"}
                assert result["issuetype"] == {"name": "Bug"}
                assert "jira_issue_url" in result

    def test_search_users(self, atlassian_tools):
        """Test search_users tool."""
        mock_users_data = [
            {
                "name": "testuser",
                "displayName": "Test User",
                "emailAddress": "test@example.com"
            }
        ]
        
        # Mock async method
        async def mock_ahttpx_get(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_users_data
            return mock_response

        with patch.object(atlassian_tools, "ahttpx_get", side_effect=mock_ahttpx_get):
            result = asyncio.run(atlassian_tools.search_users("test"))

            assert "users" in result
            assert len(result["users"]) == 1
            assert result["users"][0]["displayName"] == "Test User"
            assert result["users"][0]["emailAddress"] == "test@example.com"

    def test_jira_http_request(self, atlassian_tools):
        """Test jira_http_request tool."""
        mock_response_data = {"key": "TEST-123", "summary": "Test Issue"}

        # Mock async method
        async def mock_llm_friendly_http_request(**kwargs):
            return mock_response_data

        with patch(
            "bir_mcp.atlassian.llm_friendly_http_request",
            side_effect=mock_llm_friendly_http_request,
        ):
            result = asyncio.run(atlassian_tools.jira_http_request("issue/TEST-123"))

            assert result == mock_response_data


class TestAtlassianWriteTools(TestAtlassianTools):
    """Test write Atlassian tools."""

    def test_create_jira_issue(self, atlassian_tools):
        """Test create_jira_issue tool."""
        mock_created_issue = {"key": "TEST-456"}
        
        # Mock async method
        async def mock_ahttpx_post(endpoint, **kwargs):
            mock_response = Mock()
            mock_response.json.return_value = mock_created_issue
            return mock_response

        with patch.object(atlassian_tools, "ahttpx_post", side_effect=mock_ahttpx_post):
            result = asyncio.run(atlassian_tools.create_jira_issue(
                project_key="TEST",
                summary="New Test Issue",
                description="Test description",
                issue_type="Bug",
                assignee="testuser"
            ))

            assert result["created_issue_key"] == "TEST-456"
            assert "url" in result


class TestAtlassianHelpers(TestAtlassianTools):
    """Test Atlassian helper methods."""

    def test_build_issue_url(self, atlassian_tools):
        """Test build_issue_url helper method."""
        result = atlassian_tools.build_issue_url("TEST-123")
        assert result == "https://jira.example.com/browse/TEST-123"

    def test_prepend_api(self, atlassian_tools):
        """Test prepend_api helper method."""
        result = atlassian_tools.prepend_api("issue/TEST-123")
        assert result == "rest/api/3/issue/TEST-123"

    def test_get_tag(self, atlassian_tools):
        """Test get_tag method."""
        assert atlassian_tools.get_tag() == "atlassian"

    def test_get_mcp_tools_returns_all_tools(self, atlassian_tools):
        """Test that get_mcp_tools returns all expected tools."""
        tools = atlassian_tools.get_mcp_tools()

        # Should have both read and write tools
        assert len(tools) > 0

        # Check that specific tools are included
        tool_functions = [tool.fn for tool in tools]
        assert atlassian_tools.get_allowed_issue_types_for_project in tool_functions
        assert atlassian_tools.get_raw_jira_issue_fields in tool_functions
        assert atlassian_tools.get_jira_issue_overview in tool_functions
        assert atlassian_tools.search_users in tool_functions
        assert atlassian_tools.create_jira_issue in tool_functions
        assert atlassian_tools.jira_http_request in tool_functions

    def test_get_mcp_server_without_components(self, atlassian_tools):
        """Test get_mcp_server_without_components method."""
        server = atlassian_tools.get_mcp_server_without_components()
        assert server.name == "Bir Atlassian MCP server"
        assert "Atlassian related tools" in server.instructions


@pytest.mark.integration
class TestAtlassianIntegration:
    """Integration tests for Atlassian tools with real API calls."""

    @pytest.fixture
    def real_atlassian_tools(self):
        """Create real Atlassian tools instance for integration testing."""
        # These should be set as environment variables or skipped if not available
        import os
        jira_url = os.getenv("JIRA_URL", "https://jira-support.kapitalbank.az")
        jira_token = os.getenv("JIRA_TOKEN")
        
        if not jira_token:
            pytest.skip("JIRA_TOKEN environment variable not set")
        
        return Atlassian(
            token=jira_token,
            url=jira_url,
            api_version=3
        )

    def test_get_allowed_issue_types_integration(self, real_atlassian_tools):
        """Integration test for get_allowed_issue_types_for_project."""
        # Use real ACE project
        result = asyncio.run(real_atlassian_tools.get_allowed_issue_types_for_project("ACE"))
        
        assert "tasks" in result
        assert "subtasks" in result
        assert isinstance(result["tasks"], dict)
        assert isinstance(result["subtasks"], dict)
        
        # Should have some common issue types
        all_types = {**result["tasks"], **result["subtasks"]}
        assert len(all_types) > 0
        print(f"✅ Found {len(all_types)} issue types for ACE project")

    async def test_search_users_integration(self, real_atlassian_tools):
        """Integration test for search_users."""
        # Test multiple real user queries
        test_queries = ["amir", "farid"]
        
        for query in test_queries:
            result = await real_atlassian_tools.search_users(query)
            
            assert "users" in result
            assert isinstance(result["users"], list)
            print(f"✅ Search for '{query}' returned {len(result['users'])} users")
            
            if result["users"]:
                user = result["users"][0]
                # Check that expected fields are present
                expected_fields = ["name", "displayName", "emailAddress"]
                for field in expected_fields:
                    if field in user:  # Some fields might be missing depending on user settings
                        assert isinstance(user[field], str)
                print(f"  → First user: {user.get('displayName', 'N/A')} ({user.get('name', 'N/A')})")

    def test_jira_http_request_integration(self, real_atlassian_tools):
        """Integration test for jira_http_request."""
        # Test multiple API endpoints
        test_endpoints = [
            ("serverInfo", ["baseUrl", "version", "serverTitle"]),
            ("project/ACE", ["key", "name", "projectTypeKey"]),
            ("myself", ["accountId", "displayName"])
        ]
        
        for endpoint, expected_fields in test_endpoints:
            try:
                result = asyncio.run(real_atlassian_tools.jira_http_request(endpoint))
                
                assert isinstance(result, dict)
                print(f"✅ {endpoint}: API call successful")
                
                # Check for expected fields
                for field in expected_fields:
                    if field in result:
                        assert isinstance(result[field], str)
                        print(f"  → {field}: {result[field]}")
                        
            except Exception as e:
                print(f"⚠️  {endpoint}: {str(e)}")
                # Don't fail the test if one endpoint fails
                continue

    @pytest.mark.skip(reason="Creates real Jira issues - only run manually")
    async def test_create_jira_issue_integration(self, real_atlassian_tools):
        """Integration test for create_jira_issue (skipped by default)."""
        # This test creates a real issue, so it's skipped by default
        result = await real_atlassian_tools.create_jira_issue(
            project_key="ACE",
            summary="Integration Test Issue",
            description="This is a test issue created by integration tests",
            issue_type="Kanban Development Task"
        )
        
        assert "created_issue_key" in result
        assert "url" in result
        assert result["created_issue_key"].startswith("ACE-")
        assert "browse" in result["url"]

    def test_get_raw_jira_issue_fields_integration(self, real_atlassian_tools):
        """Integration test for get_raw_jira_issue_fields."""
        import random
        
        # Try 10 random ACE issue keys from 1-1000 range
        random_numbers = random.sample(range(1, 1001), 10)
        successful_tests = 0
        
        for num in random_numbers:
            test_issue_key = f"ACE-{num}"
            try:
                result = asyncio.run(
                    real_atlassian_tools.get_raw_jira_issue_fields(
                        test_issue_key, ["summary", "description", "status"]
                    )
                )
                
                # Should return the requested fields
                assert "summary" in result
                assert isinstance(result["summary"], str)
                print(f"✅ {test_issue_key}: {result['summary'][:50]}...")
                
                if "description" in result:
                    assert isinstance(result["description"], (str, type(None)))
                
                successful_tests += 1
                if successful_tests >= 3:  # Stop after 3 successful tests
                    break
                    
            except Exception as e:
                # If the test issue doesn't exist, try the next one
                if "does not exist" in str(e).lower() or "not found" in str(e).lower():
                    print(f"⚠️  {test_issue_key} does not exist, trying next...")
                    continue
                else:
                    raise
        
        if successful_tests == 0:
            pytest.skip("No valid ACE issues found in the tested range")
        
        print(f"✅ Successfully tested {successful_tests} ACE issues")

    def test_get_jira_issue_overview_integration(self, real_atlassian_tools):
        """Integration test for get_jira_issue_overview."""
        import random
        
        # Try 10 random ACE issue keys from 1-1000 range
        random_numbers = random.sample(range(1, 1001), 10)
        successful_tests = 0
        
        for num in random_numbers:
            test_issue_key = f"ACE-{num}"
            try:
                result = asyncio.run(real_atlassian_tools.get_jira_issue_overview(test_issue_key))
                
                # Should return formatted issue overview
                assert isinstance(result, dict)
                assert "summary" in result
                assert "jira_issue_url" in result
                assert test_issue_key in result["jira_issue_url"]
                print(f"✅ {test_issue_key} overview: {result['summary'][:50]}...")
                
                successful_tests += 1
                if successful_tests >= 2:  # Stop after 2 successful tests
                    break
                    
            except Exception as e:
                if "does not exist" in str(e).lower() or "not found" in str(e).lower():
                    print(f"⚠️  {test_issue_key} does not exist, trying next...")
                    continue
                else:
                    raise
        
        if successful_tests == 0:
            pytest.skip("No valid ACE issues found in the tested range")
        
        print(f"✅ Successfully tested {successful_tests} ACE issue overviews")



import inspect
from typing import Annotated, override

import fastmcp
import git
import pydantic

from bir_mcp.core import BaseMcp, build_readonly_tools, build_write_tools

RepoPathType = Annotated[
    str,
    pydantic.Field(
        description="The filesystem path to the local repo root or any file in the repo."
    ),
]


class Git(BaseMcp):
    @override
    def get_tag(self):
        return "git"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="MCP server with local Git tools",
            instructions=inspect.cleandoc("""
                Contains tools to work with local Git repository.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        read_tools = [
            get_local_repo_metadata,
        ]
        write_tools = [
            pull_branch,
        ]
        tools = build_readonly_tools(read_tools, max_output_length=max_output_length, tags=tags)
        tools.extend(build_write_tools(write_tools, max_output_length=max_output_length, tags=tags))
        return tools


def get_local_repo_metadata(repo_path: RepoPathType = ".") -> dict:
    """
    Retrieves metadata about the Git repo in the local filesystem, such as branch names, remote urls,
    user emails from the Git config.
    """
    repo = git.Repo(repo_path, search_parent_directories=True)
    branch_names = [b.name for b in repo.branches]
    remotes = [{"name": r.name, "urls": list(r.urls)} for r in repo.remotes]
    config = repo.config_reader()
    user_emails = config.get_values(section="user", option="email")
    metadata = {
        "remotes": remotes,
        "branches": branch_names,
        "active_branch": repo.active_branch.name,
        "user_emails": user_emails,
    }
    return metadata


def pull_branch(
    branch: Annotated[str, pydantic.Field(description="The name of the branch to checkout.")],
    remote: Annotated[
        str, pydantic.Field(description="The name of the remote to fetch from.")
    ] = "origin",
    repo_path: RepoPathType = ".",
) -> dict:
    """Pull a branch from the remote to the local Git repo."""
    repo = git.Repo(repo_path, search_parent_directories=True)
    remote: git.Remote = repo.remote(remote)
    remote.fetch()
    repo.git.checkout(branch)
    remote.pull(branch)
    response = {"status": "success"}
    return response



import inspect
from typing import override

import fastmcp
import pandas as pd
import sqlalchemy as sa

from bir_mcp.config import SqlContext
from bir_mcp.core import BaseMcp, build_readonly_tools
from bir_mcp.utils import filter_dict_by_keys, split_schema_table


class SQL(BaseMcp):
    def __init__(
        self,
        sql_context: SqlContext,
        sample_rows_in_table_info: int = 5,
        timezone: str = "UTC",
    ):
        super().__init__(timezone=timezone)
        self.sql_context = sql_context
        self.engine = self.sql_context.connection.get_engine()
        self.inspector = sa.inspect(self.engine)
        self.sample_rows_in_table_info = sample_rows_in_table_info

    @override
    def get_tag(self):
        name = self.sql_context.name or self.sql_context.connection_name
        return f"sql_{name}"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="MCP server with SQL tools",
            instructions=inspect.cleandoc("""
                Contains tools to work with SQL databases.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        read_tools = [
            self.get_database_info,
            self.get_available_table_names,
            self.get_table_info,
            self.execute_query,
        ]
        tools = build_readonly_tools(read_tools, max_output_length=max_output_length, tags=tags)
        return tools

    def read_sql_to_dict(self, sql) -> dict:
        df = pd.read_sql(sql, self.engine)
        df_dict = df.to_dict(orient="split", index=False)
        # The CSV option if more reliable in terms of being able to serialize to string.
        # df = pd.read_sql(sql, self.engine)
        # buffer = io.StringIO()
        # df.to_csv(buffer, index=False)
        # csv_table = buffer.getvalue()
        return df_dict

    def get_database_info(self) -> dict:
        """
        Get info about the database, such as its dialect and version,
        as well as the connection name for reference in other tools.
        """
        dialect = self.inspector.dialect
        info = {
            "dialect": dialect.name,
            "connection_name": self.sql_context.connection_name,
        }
        if dialect.server_version_info:
            info["server_version"] = ".".join(map(str, dialect.server_version_info))

        return info

    def get_available_table_names(self) -> dict:
        """Get available schema-qualified table names."""
        # The available tables are those that are explicitly specified in the config.
        # It is done this way for predictability and efficiency to avoid fetching all
        # table names from the database, which can be slow for large databases.
        # It is not guaranteed that SQL connection user has permissions to select the
        # returned tables, but to check it, either a fragile probe SQL is needed or
        # a table reflection has to be performed, which can add up to the latency.
        schema_tables = []
        for schema_table in self.sql_context.schema_tables:
            schema, table = split_schema_table(schema_table)
            if self.inspector.has_table(schema=schema, table_name=table):
                schema_tables.append(schema_table)

        tables = {"schema_tables": sorted(schema_tables)}
        return tables

    def get_table_info(self, schema_table: str, include_sample_rows: bool = False) -> dict:
        """
        Get SQL table info, such as the create table SQL statement, column comments,
        primary and foreign keys, indexes, and sample rows.
        """
        schema_tables = self.get_available_table_names()["schema_tables"]
        if schema_table not in schema_tables:
            raise ValueError(f"Table {schema_table} is not available.")

        schema, table = split_schema_table(schema_table)

        sa_table = sa.Table(table, sa.MetaData(), schema=schema, autoload_with=self.engine)
        create_table = sa.schema.CreateTable(sa_table).compile(self.engine)
        create_table = "\n".join(line.strip() for line in str(create_table).strip().splitlines())

        column_comments = {c.name: c.comment for c in sa_table.columns if c.comment}

        primary_key = self.inspector.get_pk_constraint(table, schema)
        primary_key_columns = primary_key.get("constrained_columns")

        foreign_keys = self.inspector.get_foreign_keys(table, schema)
        foreign_keys = [
            filter_dict_by_keys(
                fk, ["constrained_columns", "referred_schema", "referred_table", "referred_columns"]
            )
            for fk in foreign_keys
        ]

        indexes = self.inspector.get_indexes(table, schema)
        index_columns = [i.get("column_names") for i in indexes]

        table_info = {
            "create_table_statement": create_table,
            "column_comments": column_comments,
            "primary_key_columns": primary_key_columns,
            "foreign_keys": foreign_keys,
            "index_columns": index_columns,
        }
        if include_sample_rows:
            query = sa.select(sa_table).limit(self.sample_rows_in_table_info)
            sample_rows = self.read_sql_to_dict(query)
            table_info["sample_rows"] = sample_rows

        return table_info

    def execute_query(self, query: str) -> dict:
        """
        Execute an SQL query and return the result. Whenever possible, follow the following optimization principles:
        - Select only the columns that are needed.
        - Use aggregations to reduce the amount of data returned.
        - When joining, filtering and sorting, prefer indexed columns.
        - Avoid applying functions to indexed columns in the filtering clauses.
        """
        table = self.read_sql_to_dict(query)
        return table



import datetime
import enum
import inspect
from typing import Annotated, override

import fastmcp.prompts
import fastmcp.resources
import fastmcp.tools
import httpx
import pydantic

from bir_mcp.core import BaseMcp, build_readonly_tools
from bir_mcp.grafana.prompts import get_prompts
from bir_mcp.grafana.resources import get_mcp_resource_uri_functions
from bir_mcp.grafana.utils import to_grafana_time_format
from bir_mcp.utils import (
    araise_for_status,
    filter_dict_by_keys,
    llm_friendly_http_request,
    prepend_url_path_prefix_if_not_present,
    to_maybe_ssl_context,
)


class DatasourceType(enum.StrEnum):
    alertmanager = "Alertmanager"
    clickhouse = "ClickHouse"
    grafana_pyroscope = "Grafana Pyroscope"
    influxdb = "InfluxDB"
    json_api = "JSON API"
    loki = "Loki"
    microsoft_sql_server = "Microsoft SQL Server"
    oracle_enterprise_manager_data_source = "Oracle Enterprise Manager Data Source"
    postgresql = "PostgreSQL"
    prometheus = "Prometheus"
    tempo = "Tempo"
    zabbix = "Zabbix"


class InfluxDbQueryType(enum.StrEnum):
    flux = enum.auto()
    influxql = enum.auto()


class Grafana(BaseMcp):
    def __init__(
        self,
        token_or_auth: str | tuple[str, str],
        url: str = "https://yuno.kapitalbank.az",
        http_timeout_seconds: int = 30,
        timezone: str = "UTC",
        ssl_verify: bool | str = True,
    ):
        super().__init__(timezone=timezone)
        if isinstance(token_or_auth, str):
            headers = {"Authorization": f"Bearer {token_or_auth}"}
            auth = None
        else:
            headers = {}
            auth = token_or_auth

        self.ahttpx = httpx.AsyncClient(
            auth=auth,
            headers=headers,
            base_url=url,
            event_hooks={"response": [araise_for_status]},
            timeout=http_timeout_seconds,
            verify=to_maybe_ssl_context(ssl_verify),
        )

    @override
    def get_tag(self):
        return "grafana"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="Bir Grafana MCP server",
            instructions=inspect.cleandoc("""
                Use this server for tasks related to Grafana and its datasources.
                Note that in context of the bank, Grafana may be referred to as "Yuno",
                this is due to the fact that the domain name for the on-prem Grafana instance is "yuno.kapitalbank.az".
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        functions = [
            self.list_all_datasources,
            self.query_influxdb_datasource,
            self.query_datasource,
            self.http_request,
        ]
        tools = build_readonly_tools(functions, max_output_length=max_output_length)
        return tools

    @override
    def get_uri_resources(self):
        return get_mcp_resource_uri_functions()

    @override
    def get_prompts(self):
        return get_prompts()

    def prepend_api_version(self, path: str) -> str:
        return prepend_url_path_prefix_if_not_present("api", path)

    async def http_request(
        self,
        path: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The path to the Grafana API endpoint, relative to the base URL.
                    Should not include the /api prefix.
                """)
            ),
        ],
        http_method: str = "GET",
        json_query_params: Annotated[
            str | None,
            pydantic.Field(description="The url query parameters dict, serialized to JSON."),
        ] = None,
        content: Annotated[
            str | dict | None,
            pydantic.Field(description="The request body."),
        ] = None,
    ) -> dict:
        """
        Performs a HTTP request to the Grafana REST API. Can be used when no specialized tool is available for the task.
        The arguments are passed to a httpx.Client.request method, with base_url set to local
        Grafana instance and API key set in headers, so the result looks like this:
        return client.request(method=http_method, url=path, params=json.loads(json_query_params), content=content).json()
        """
        data = await llm_friendly_http_request(
            client=self.ahttpx,
            url=self.prepend_api_version(path),
            http_method=http_method,
            json_query_params=json_query_params,
            content=content,
            ensure_dict_output=True,
        )
        return data

    async def list_all_datasources(self) -> dict:
        """Lists all datasources in the Grafana instance."""
        response = await self.ahttpx.get("/api/datasources")
        datasources = {
            "datasources": [
                filter_dict_by_keys(d, ["uid", "name", "typeName"]) for d in response.json()
            ]
        }
        return datasources

    async def query_influxdb_datasource(
        self,
        datasource_uid: Annotated[
            str,
            pydantic.Field(description="The uid of an InfluxDB Grafana datasource."),
        ],
        query: Annotated[
            str,
            pydantic.Field(description="The query to execute."),
        ],
        start_time: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The start time for the query in relative Grafana time units like "now", "now-5m", 
                    "now-1h", "now-7d", or in ISO 8601 format "YYYY-MM-DDTHH:MM:SS+HH:MM".
                    Can be refered to as "v.timeRangeStart" in queries.
                    Should not be older than one month for query efficiency and because data is probably 
                    not retained longer than that anyway.
                """)
            ),
        ] = "now-1h",
        end_time: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The end time for the query, in same format as start_time. 
                    Can be refered to as "v.timeRangeStop" in queries.
                """)
            ),
        ] = "now",
        query_language: Annotated[
            InfluxDbQueryType,
            pydantic.Field(
                description=inspect.cleandoc(f"""
                The language of query to execute, {InfluxDbQueryType.influxql} for SQL-like Influx query language, 
                and {InfluxDbQueryType.flux} for Flux functional data scripting language.
            """)
            ),
        ] = InfluxDbQueryType.flux,
    ) -> dict:
        """
        Executes a query against an InfluxDB datasource in the Grafana instance.
        Note that InfluxQL may not be supported by the InfluxDB datasource.
        Refer to Flux instruction in the MCP resources for details about the Flux language.
        This function delegates the call to the query_datasource function,
        so refer to its documentation for additional details.
        """
        tables = await self.query_datasource(
            datasource_uid=datasource_uid,
            query=query,
            additional_datasource_params={"queryType": query_language.value},
            start_time=start_time,
            end_time=end_time,
        )
        now = datetime.datetime.now()
        min_timestamp = int((now - datetime.timedelta(days=1000)).timestamp() * 1000)
        max_timestamp = int((now + datetime.timedelta(minutes=1)).timestamp() * 1000)
        for table in tables["tables"]:
            table["fields"] = [filter_dict_by_keys(f, ["name", "labels"]) for f in table["fields"]]
            data = []
            for column in table["data"]:
                # Convert millisecond timestamps to datetime strings.
                if any(column) and all(
                    i is None or isinstance(i, int) and min_timestamp < i < max_timestamp
                    for i in column
                ):
                    column = [datetime.datetime.fromtimestamp(i / 1000) if i else i for i in column]
                    column = [self.format_datetime_for_ai(i) for i in column]

                data.append(column)

            table["data"] = data

        return tables

    async def query_datasource(
        self,
        datasource_uid: Annotated[
            str,
            pydantic.Field(description="The uid of a Grafana datasource."),
        ],
        query: Annotated[
            str,
            pydantic.Field(description="The query to execute."),
        ],
        additional_datasource_params: Annotated[
            dict[str, str] | None,
            pydantic.Field(
                description="Additional datasource parameters, for example queryType for InfluxDB queries."
            ),
        ] = None,
        start_time: Annotated[
            str,
            pydantic.Field(
                description=inspect.cleandoc("""
                    The start time for the query in relative Grafana time units like "now", "now-5m", 
                    "now-1h", "now-7d", or in ISO 8601 format "YYYY-MM-DDTHH:MM:SS+HH:MM".
                    Should not be older than one month for query efficiency and because data is probably 
                    not retained longer than that anyway.
                """)
            ),
        ] = "now-1h",
        end_time: Annotated[
            str,
            pydantic.Field(description="The end time for the query, in same format as start_time."),
        ] = "now",
    ) -> dict:
        """
        Executes a query against a datasource in the Grafana instance.
        If query returns too much data, it will lead to HTTP timeouts.
        To avoid this, the start to end time range should be as small as possible,
        and the query should aggregate data by time windows.
        Refer to [Grafana endpoint docs](https://grafana.com/docs/grafana/latest/developers/http_api/data_source/#query-a-data-source)
        for details about the used endpoint.
        """
        ref_id = "A"
        datasource_query = (additional_datasource_params or {}) | {
            "datasource": {"uid": datasource_uid},
            "refId": ref_id,
            "format": "table",
            "maxDataPoints": 100,
            "intervalMs": 1000,
        }
        datasources = await self.list_all_datasources()
        datasource_types = {d["uid"]: d["typeName"] for d in datasources["datasources"]}
        datasource_type = datasource_types.get(datasource_uid)
        match datasource_type:
            case "PostgreSQL":
                datasource_query["rawSql"] = query
            case _:
                datasource_query["query"] = query

        payload = {
            "queries": [datasource_query],
            "from": to_grafana_time_format(start_time),
            "to": to_grafana_time_format(end_time),
        }
        response = await self.ahttpx.post("/api/ds/query", json=payload)
        tables = []
        result = response.json()["results"].get(ref_id)
        frames = result["frames"] if result else []
        for frame in frames:
            table = filter_dict_by_keys(frame["schema"], ["name", "fields"])
            table["data"] = frame["data"]["values"]
            tables.append(table)

        tables = {"tables": tables}
        return tables


# def build_prometheus_from_grafana_datasource_proxy(
#     grafana_url: str, datasource: str, auth: tuple[str, str], **kwargs
# ):
#     """https://grafana.com/docs/grafana/v11.2/developers/http_api/data_source/#data-source-proxy-calls-by-id"""
#     data_sources = fetch_grafana_datasources(grafana_url, auth)
#     data_source = data_sources.filter(pl.col("name") == datasource)
#     if len(data_source) != 1:
#         raise ValueError(f"Couldn't find a singular id for data source {data_source}.")

#     data_source_id = data_source["id"].item()
#     prometheus_url = join_url_components(grafana_url, "api", "datasources", "proxy", data_source_id)
#     prometheus = Prometheus(prometheus_url, auth=auth, **kwargs)
#     return prometheus


# class Prometheus:
#     """https://prometheus.io/docs/prometheus/latest/querying/api"""

#     def __init__(
#         self,
#         url: str,
#         api_version: str = "v1",
#         default_start=datetime.timedelta(days=-1),
#         default_end=None,
#         time_zone: str = "Asia/Baku",
#         auth=None,
#     ):
#         self.url = url
#         self.api_version = api_version
#         self.default_date_keys = {
#             "start": default_start,
#             "end": default_end,
#             "time": default_end,
#         }
#         self.time_zone = time_zone
#         self.auth = auth

#     def request_as_dict(self, *url_components, **params) -> dict:
#         params = {k: v for k, v in params.items() if v}
#         match url_components:
#             case (
#                 ["query"]
#                 | ["query_range"]
#                 | ["series"]
#                 | ["labels"]
#                 | ["label", _, "values"]
#                 | ["query_exemplars"]
#             ):
#                 for key, default in self.default_date_keys.items():
#                     value = params.get(key, default)
#                     params[key] = to_datetime(value).isoformat()

#         url = join_url_components(self.url, "api", self.api_version, *url_components)
#         response_dict = request_as_dict(url, params=params, auth=self.auth)
#         if response_dict["status"] == "error":
#             raise RuntimeError(
#                 f'Prometheus API request for "{join_url_components(*url_components)}" endpoint '
#                 f"with parameters {params} returned an error:\n"
#                 f"{response_dict}"
#             )

#         data = response_dict["data"]
#         return data

#     def response_metric_data_to_dataframe(self, data: dict) -> pl.DataFrame | None:
#         if not (result := data["result"]):
#             return

#         rows = []
#         for series in result:
#             match data["resultType"]:
#                 case "matrix":
#                     values = series["values"]
#                 case "vector":
#                     values = [series["value"]]

#             for timestamp, value in values:
#                 row = series["metric"] | {"timestamp": timestamp, "value": value}
#                 rows.append(row)

#         if not rows:
#             return

#         df = pl.DataFrame(rows)
#         df = df.cast({"value": pl.Float64})
#         df = df.with_columns(pl.from_epoch("timestamp").dt.convert_time_zone(self.time_zone))
#         return df

#     def query(self, query, time=None, timeout: str | None = None) -> pl.DataFrame | None:
#         """https://prometheus.io/docs/prometheus/latest/querying/api/#instant-queries"""
#         query = re.sub(r"\s+", " ", query)
#         data = self.request_as_dict("query", query=query, time=time, timeout=timeout)
#         df = self.response_metric_data_to_dataframe(data)
#         return df

#     def query_range(
#         self,
#         query: str,
#         start=None,
#         end=None,
#         step: str | float = "1m",
#         timeout: str | None = None,
#     ) -> pl.DataFrame | None:
#         """https://prometheus.io/docs/prometheus/latest/querying/api/#range-queries"""
#         data = self.request_as_dict(
#             "query_range",
#             query=re.sub(r"\s+", " ", query),
#             start=start,
#             end=end,
#             step=step,
#             timeout=timeout,
#         )
#         df = self.response_metric_data_to_dataframe(data)
#         return df

#     def get_all_metric_names(self, **kwargs) -> list[str]:
#         """https://prometheus.io/docs/prometheus/latest/querying/api/#querying-label-values"""
#         data = self.request_as_dict("label", "__name__", "values", **kwargs)
#         return data

#     def get_time_series(self, series_selector: str, start=None, end=None) -> pl.DataFrame | None:
#         """https://prometheus.io/docs/prometheus/latest/querying/api/#finding-series-by-label-matchers"""
#         data = self.request_as_dict(
#             "series", **{"match[]": series_selector, "start": start, "end": end}
#         )
#         if not data:
#             return

#         df = pl.DataFrame(data)
#         return df



import inspect
import tempfile
from typing import Annotated, override

import detect_secrets
import fastmcp
import pydantic

from bir_mcp.core import BaseMcp, build_readonly_tools


class Filesystem(BaseMcp):
    @override
    def get_tag(self):
        return "filesystem"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="MCP server with local filesystem tools",
            instructions=inspect.cleandoc("""
                Contains tools to work with local files.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        read_tools = [
            find_secrets,
            find_secrets_in_text,
        ]
        tools = build_readonly_tools(
            read_tools, max_output_length=max_output_length, tags={self.get_tag()}
        )
        return tools


def find_secrets(
    file_path: Annotated[
        str,
        pydantic.Field(description="The path to the file to scan for secrets."),
    ],
) -> dict:
    """
    Scans a file for secrets using the [detect-secrets](https://github.com/Yelp/detect-secrets) tool.
    """
    secrets = detect_secrets.SecretsCollection()
    with detect_secrets.settings.default_settings():
        secrets.scan_file(file_path)

    secrets = [
        {
            "secret_type": secret.type,
            "line_number": secret.line_number,
        }
        for file_secrets in secrets.data.values()
        for secret in file_secrets
    ]
    secrets = {"detected_secrets": secrets}
    return secrets


def find_secrets_in_text(text: str) -> dict:
    """Scan text content for secrets."""
    with tempfile.NamedTemporaryFile() as temp_file:
        temp_file.write(text.encode())
        temp_file.flush()
        secrets = find_secrets(temp_file.name)

    return secrets



import datetime
import enum
import inspect
from typing import Annotated, override

import fastmcp.prompts
import fastmcp.resources
import fastmcp.tools
import httpx
import pydantic

from bir_mcp.core import BaseMcp, build_readonly_tools
from bir_mcp.grafana.prompts import get_prompts
from bir_mcp.grafana.resources import get_mcp_resource_uri_functions
from bir_mcp.grafana.utils import to_grafana_time_format
from bir_mcp.utils import araise_for_status, filter_dict_by_keys, to_maybe_ssl_context


class Dataiku(BaseMcp):
    def __init__(
        self,
        auth: tuple[str, str],
        url: str = "https://dssdesign-data.kapitalbank.az",
        http_timeout_seconds: int = 30,
        timezone: str = "UTC",
        ssl_verify: bool | str = True,
    ):
        super().__init__(timezone=timezone)
        self.client = httpx.AsyncClient(
            auth=auth,
            base_url=url,
            event_hooks={"response": [araise_for_status]},
            timeout=http_timeout_seconds,
            verify=to_maybe_ssl_context(ssl_verify),
        )

    @override
    def get_tag(self):
        return "dataiku"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="Bir Dataiku MCP server",
            instructions=inspect.cleandoc("""
                Use this server for tasks related to Dataiku.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        functions = []
        tools = build_readonly_tools(functions, max_output_length=max_output_length)
        return tools

    async def get_project_list(self) -> dict: ...



import bidict


def get_mcp_resource_uri_functions() -> bidict.bidict[str, callable]:
    functions = bidict.bidict(
        {
            "instruction://flux": get_flux_instruction,
        }
    )
    return functions


def get_flux_instruction() -> str:
    """Provides information about InfluxDB Flux language, such as its data model, syntax, and usage examples."""
    instruction = """
    # Flux 
    Flux is a functional data scripting language designed for querying, analyzing, and acting on time-series data in InfluxDB.

    ## Data model
    Flux organizes data into independent buckets, similar to logical databases in SQL.
    Each bucket contains time series data. A single point is composed of:
    - measurement: The name of the thing you are measuring (e.g., "cpu", "http_requests"), similar to an SQL table or Prometheus metric name.
    - tags: Key-value pairs of metadata that describe the data (e.g., `host="server1"`, `region="us-west"`), similar to Prometheus labels.
    - fields: The actual data values (e.g., `usage_percent=99.5`, `request_count=124`), similar to Prometheus metric value.
    - timestamp: The time of the data point.

    ## Syntax
    Flux queries consist of sequentially executed pipeline of functions, joined by the pipe-forward operator "|>".
    It takes the output of one function and sends it as the input to the next function.

    ### Main Flux functions
    - from() - Specifies the InfluxDB bucket to retrieve data from. It's the starting point for query that fetches time-series data.
    - range() - Filters data based on a time range. Mandatory for time-series queries.
    - filter() - Filters rows based on column values (measurement, field, tags).
    - pivot() - Rotates data from a tall format (one row per timestamp/metric) to a wide format.
    - keep() / drop() - Filters data by column names, allowing you to keep or discard specific columns.
    - limit() - Restricts the number of rows returned.
    - group() - Groups rows together based on common column values for aggregation.
    - aggregateWindow() - Segments data into time windows and applies an aggregate function (mean, sum, etc.).
    - map() - Applies a custom function to each row to modify or add columns.
    - mean(), sum(), count(), last() - Common aggregate functions, often used inside aggregateWindow.
    - highestAverage(), highestMax(), lowestAverage(), top() - Efficient "top-n" results from a group, are significantly more performant for datasets with high tag cardinality
    - yield() - Specifies a result set to be delivered from the query.

    ### Meta functions
    - buckets() - Returns a list of all available buckets.
    - schema.measurements() - Returns a list of all measurements within a bucket.
    - schema.tagKeys() / schema.tagValues() - Returns a list of tag keys or tag values for a given measurement.

    ## Usage
    When using Flux through the Grafana's `/api/ds/query` endpoint, several variables are injected into the query,
    the most important are the `v.timeRangeStart` and `v.timeRangeStop`, which correspond to the `start_time` and `end_time` parameters
    passed to the `/api/ds/query` endpoint. So each query that fetches data from time series should look like this:

    ```flux
    // Specify the bucket name here
    from(bucket: "{bucket}")
        |> range(start: v.timeRangeStart, stop: v.timeRangeStop)
        // Filters, transformations, aggegations, etc.
        |> yield(name: "value")
    ```

    ### Examples

    #### Get all buckets
    ```flux
    buckets()
    ```

    #### Get all measurements in a given bucket
    ```flux
    // Import the schema package to access its functions
    import "influxdata/influxdb/schema"
    // Call the measurements function, specifying which bucket to look in
    schema.measurements(bucket: "{bucket}")
    ```

    #### Get all field keys for a given measurement
    ```flux
    import "influxdata/influxdb/schema"
    schema.measurementFieldKeys(
        bucket: "{bucket}",
        measurement: "{measurement}"
    )
    ```

    #### Get all tag keys for a given measurement
    ```flux
    import "influxdata/influxdb/schema"
    schema.measurementTagKeys(
        bucket: "{bucket}",
        measurement: "{measurement}"
    )
    ```

    #### Get all tag values for a given tag key
    ```flux
    import "influxdata/influxdb/schema"
    schema.measurementTagValues(
        bucket: "{bucket}",
        measurement: "{measurement}",
        tag: "{tag}"
    )
    ```

    #### Filtering and aggregation
    ```flux
    from(bucket: "{bucket}")
        |> range(start: v.timeRangeStart, stop: v.timeRangeStop)
        |> filter(fn: (r) => r._measurement == "{measurement}" and r._field == "{field}")
        // Pivot makes each unique field a new column, with timestamps as rows.
        // This is useful for creating tables or graphs with multiple series.
        |> pivot(rowKey:["_time"], columnKey: ["_field"], valueColumn: "_value")
        // Group data into 10-minute windows and calculate the mean for each window
        |> aggregateWindow(every: 10m, fn: mean)
        |> yield(name: "mean_{field}_10m")
    ```

    #### Finding lowest average values among groups
    ```flux
    from(bucket: "{bucket}")
        |> range(start: v.timeRangeStart, stop: v.timeRangeStop)
        |> filter(fn: (r) => r._measurement == "{measurement}" and r._field == "{field}")
        |> lowestAverage(n: {n}, groupColumns: ["{tag}"])
    ```
    """
    return instruction



import enum
from typing import Self, override

import pydantic
import sqlalchemy as sa
import yaml

from bir_mcp.hashicorp import ConsulKeyValue
from bir_mcp.utils import set_ssl_cert_file_from_cadata


class Driver(enum.StrEnum):
    oracle = enum.auto()
    postgresql = enum.auto()
    mysql = enum.auto()

    @property
    def name_and_package(self):
        match self:
            case Driver.oracle:
                return "oracle+oracledb"
            case Driver.postgresql:
                return "postgresql+psycopg"
            case Driver.mysql:
                return "mysql+mysqlconnector"

        assert False


class SqlConnection(pydantic.BaseModel):
    driver: Driver
    host: str
    port: int
    database: str | None = None
    username: str
    password: str
    parameters: dict[str, str] | None = None

    @property
    def url(self) -> sa.URL:
        url = sa.URL.create(
            drivername=self.driver.name_and_package,
            host=self.host,
            port=self.port,
            username=self.username,
            password=self.password,
            database=self.database,
            query=self.parameters or {},
        )
        return url

    def get_engine(self) -> sa.Engine:
        engine = sa.create_engine(self.url)
        return engine


class SqlContext(pydantic.BaseModel):
    # May need to remove this line for case sensitive table names.
    model_config = pydantic.ConfigDict(str_to_lower=True)

    name: str | None = None
    connection_name: str | None = None
    connection: SqlConnection | None = None
    schema_tables: list[str] = []


class SystemManagedConfig(pydantic.BaseModel):
    gitlab_private_token: str | None = None
    grafana_username: str | None = None
    grafana_password: str | None = None
    grafana_token: str | None = None
    jira_token: str | None = None
    gitlab_url: str = "https://gitlab.kapitalbank.az"
    grafana_url: str = "https://yuno.kapitalbank.az"
    jira_url: str = "https://jira-support.kapitalbank.az"
    sonarqube_token: str | None = None
    sonarqube_url: str = "https://sonarqube.kapitalbank.az"
    ca_file: str | None = None
    sql_connections: dict[str, SqlConnection] = {}
    sql_contexts: list[SqlContext] = []

    def get_grafana_token_or_auth(self) -> str | tuple[str, str] | None:
        if self.grafana_token:
            return self.grafana_token

        if self.grafana_username and self.grafana_password:
            return self.grafana_username, self.grafana_password

        return None

    @classmethod
    def from_consul(
        cls, consul_host: str, consul_key: str, consul_token: str | None = None
    ) -> Self:
        consul_key_value = ConsulKeyValue(host=consul_host, key=consul_key, token=consul_token)
        value = consul_key_value.load()
        config = yaml.safe_load(value)
        config = cls(**config)
        return config

    @override
    def model_post_init(self, context) -> None:
        if self.ca_file:
            set_ssl_cert_file_from_cadata(self.ca_file)

        for sql_context in self.sql_contexts:
            if not sql_context.connection:
                sql_context.connection = self.sql_connections[sql_context.connection_name]



import datetime
import inspect
from typing import override

import fastmcp.tools
import httpx

from bir_mcp.core import AhttpxMixin, BaseMcp, build_readonly_tools
from bir_mcp.git_lab.annotations import ProjectPathOrUrlType
from bir_mcp.git_lab.utils import GitLabUrl
from bir_mcp.utils import araise_for_status, build_url, filter_dict_by_keys, to_maybe_ssl_context


class SonarQube(BaseMcp, AhttpxMixin):
    def __init__(
        self,
        token: str,
        url: str = "https://sonarqube.kapitalbank.az",
        gitlab_url: str = "https://gitlab.kapitalbank.az",
        ssl_verify: bool | str = True,
        timezone: str = "UTC",
    ):
        super().__init__(timezone=timezone)
        self.ahttpx = httpx.AsyncClient(
            base_url=url,
            auth=(token, ""),
            verify=to_maybe_ssl_context(ssl_verify),
            event_hooks={"response": [araise_for_status]},
        )
        self.gitlab_url = GitLabUrl(base_url=gitlab_url)
        self.main_metrics = [
            "alert_status",
            "reliability_rating",
            "new_reliability_rating",
            "security_rating",
            "new_security_rating",
            "sqale_rating",
            "new_maintainability_rating",
            "security_review_rating",
            "new_security_review_rating",
            "coverage",
            "new_coverage",
            "duplicated_lines_density",
            "new_duplicated_lines_density",
            "sqale_index",
            "bugs",
            "new_bugs",
            "vulnerabilities",
            "new_vulnerabilities",
            "code_smells",
            "new_code_smells",
            "security_hotspots",
            "new_security_hotspots",
            "security_hotspots_reviewed",
            "new_security_hotspots_reviewed",
            "complexity",
            "cognitive_complexity",
        ]

    @override
    def get_tag(self):
        return "sonarqube"

    @override
    def get_mcp_server_without_components(self):
        server = fastmcp.FastMCP(
            name="Bir SonarQube MCP server",
            instructions=inspect.cleandoc("""
                SonarQube related tools.
            """),
        )
        return server

    @override
    def get_mcp_tools(self, max_output_length: int | None = None, tags: set[str] | None = None):
        read_tools = [
            self.get_main_project_metrics,
            self.get_project_quality_gates_status,
            self.http_request,
        ]
        tools = build_readonly_tools(read_tools, max_output_length=max_output_length, tags=tags)
        return tools

    async def list_all_metrics(self, page_size: int = 500) -> list[dict]:
        response = await self.ahttpx_get("metrics/search", ps=page_size)
        metrics = response.json()["metrics"]
        metrics = [filter_dict_by_keys(metric, ["key", "description"]) for metric in metrics]
        return metrics

    def gitlab_to_sonarqube_project(self, project_path_or_url: ProjectPathOrUrlType) -> str:
        project_path = self.gitlab_url.extract_project_path(project_path_or_url)
        # The convention is that SonarQube projects are named as "{gitlab_group}-{gitlab_project}".
        sonarqube_project = "-".join(project_path.split("/")[-2:]).lower()
        return sonarqube_project

    async def get_analysis_age(self, project: str, branch: str = "master") -> datetime.timedelta:
        response = await self.ahttpx_get("/api/components/show", component=project, branch=branch)
        analysis_date = response.json()["component"]["analysisDate"]
        age = datetime.datetime.now(datetime.UTC) - datetime.datetime.fromisoformat(analysis_date)
        return age

    async def get_main_project_metrics(
        self, project_path_or_url: ProjectPathOrUrlType, branch: str = "master"
    ) -> dict:
        """
        Get the main metrics for a project, such as alert status, reliability and security ratings,
        coverage, bugs, complexity, etc.
        """
        project = self.gitlab_to_sonarqube_project(project_path_or_url)
        response = await self.ahttpx_get(
            "measures/component",
            component=project,
            branch=branch,
            metricKeys=",".join(self.main_metrics),
        )
        measures = response.json()["component"]["measures"]
        metrics = []
        for measure in measures:
            if period := measure.pop("period", None):
                measure.pop("periods")
                period.pop("index")
                measure |= period

            metrics.append(measure)

        age = await self.get_analysis_age(project=project, branch=branch)
        url = build_url(self.ahttpx.base_url, "dashboard", id=project, branch=branch)
        metrics = {
            "metrics": metrics,
            "analysis_age": str(age),
            "sonarqube_dashboard_url": url,
        }
        return metrics

    async def get_project_quality_gates_status(
        self, project_path_or_url: ProjectPathOrUrlType, branch: str = "master"
    ) -> dict:
        """Get the quality gates status for a project."""
        project = self.gitlab_to_sonarqube_project(project_path_or_url)
        response = await self.ahttpx_get(
            "qualitygates/project_status", projectKey=project, branch=branch
        )
        project_status = response.json()["projectStatus"]
        conditions = [
            filter_dict_by_keys(i, ["metricKey", "status", "actualValue", "errorThreshold"])
            for i in project_status["conditions"]
        ]
        age = await self.get_analysis_age(project=project, branch=branch)
        project_status = {
            "status": project_status["status"],
            "conditions": conditions,
            "analysis_age": str(age),
        }
        return project_status
